from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    pass

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)
handler = logging.StreamHandler()
handler.setLevel(logging.WARNING)
logger.addHandler(handler)

__author__ = "Felix Jung and Debsankha Manik"
__email__ = "felix.jung@tu-dresden.de; dmanik@debsankha.net"



import typer
import gnupg
import os
import re
import subprocess
import tomli
import tomli_w
import packaging.version

try:
    # only relevant for dev mode
    import pygit2
except ImportError:
    pass

import pandas as pd

from copy import deepcopy
from pathlib import Path
from typing import Optional, List, Union, Annotated
from pygit2 import (
    GIT_STATUS_WT_MODIFIED,
    GIT_STATUS_INDEX_MODIFIED,
    GIT_MERGE_ANALYSIS_UP_TO_DATE,
)
from enum import Enum

from ridepy.extras.io import read_params_json
from ridepy.extras.io_utils import (
    update_filenames as _update_filenames,
    update_events_files as _update_events_files,
)

from ridepy.extras.simulation_set import (
    perform_single_simulation,
    perform_single_analysis,
)

app = typer.Typer()

hpc_app = typer.Typer()
app.add_typer(hpc_app, name="hpc")

dev_app = typer.Typer()
app.add_typer(dev_app, name="dev")


@app.command()
def update_filenames(
    directory: Optional[Path] = typer.Argument(
        None, dir_okay=True, file_okay=False, exists=True
    )
):
    """
    Update filenames and JSON structure to up-to-date simulation ids.

    This reads all `.json` files, tries to parse the possibly
    outdated format and generates a simulation id according
    to the current scheme. Afterwards, new JSON files are stored using the
    current format at `<new id>.json`.
    The simulation data files ending with `.jsonl` are not modified. Instead,
    symbolic links are created, pointing from `<new id>.jsonl` to `<old id>.jsonl`.
    """
    if directory is None:
        directory = Path.cwd()
    else:
        directory = Path(directory).expanduser()

    _update_filenames(directory)


@app.command()
def update_events_files(
    directory: Optional[Path] = typer.Option(
        None, dir_okay=True, file_okay=False, exists=True
    ),
):
    """
    Update events files (*.jsonl) to current, flattened structure.
    """
    if directory is None:
        directory = Path.cwd()
    else:
        directory = Path(directory).expanduser()

    _update_events_files(directory)


@hpc_app.command()
def simulate(
    input_directory: Path = typer.Argument(
        ...,
        help="Directory containing the parameter files for the simulation",
        dir_okay=True,
        file_okay=False,
        exists=True,
        readable=True,
        resolve_path=True,
    ),
    output_directory: Path = typer.Argument(
        ...,
        help="Directory to store the simulation results in",
        dir_okay=True,
        file_okay=False,
        writable=True,
        resolve_path=True,
    ),
    simulation_id: str = typer.Argument(..., help="ID of a single simulation."),
):
    """
    Perform a single simulation run reading the simulation
    parameters from a file and writing the output to disk.

    Which file the parameters are read from is determined by
    the SIMULATION_ID argument.

    The parameter files must be located in the INPUT_DIRECTORY,
    and be named like "<SIMULATION_ID>_params.json"
    """

    params_path = input_directory / f"{simulation_id}_params.json"

    params_json = read_params_json(params_path)
    perform_single_simulation(params=params_json, data_dir=output_directory)


@hpc_app.command()
def analyze(
    output_directory: Path = typer.Argument(
        ...,
        help="Directory to read the simulation results in",
        dir_okay=True,
        file_okay=False,
        readable=True,
        writable=True,
        resolve_path=True,
    ),
    simulation_ids: List[str] = typer.Argument(..., help="One or many simulation IDs."),
    compute_system_quantities: bool = typer.Option(
        False, help="Compute system quantities"
    ),
):
    """
    Analyze a single simulation run. This reads the simulation data
    from the params and event files and writes the output to disk
    in the form of files in parquet format.

    Which file the simulation results and parameters are read from
    is determined by the SIMULATION_ID argument.

    The result and parameter files must be located in the OUTPUT_DIRECTORY,
    and be named like "<SIMULATION_ID>.jsonl" and "<SIMULATION_ID>_params.json",
    respectively.
    """

    tasks = {"stops", "requests", "vehicle_quantities"}
    if compute_system_quantities:
        tasks.add("system_quantities")

    system_quantities = []
    for simulation_id in simulation_ids:
        _, sys_quant = perform_single_analysis(
            sim_id=simulation_id,
            data_dir=output_directory,
            tasks_if_existent=tasks,
        )
        system_quantities.append(sys_quant)

    if compute_system_quantities:
        system_quantities_df = pd.DataFrame(system_quantities, index=simulation_ids)
        system_quantities_df.rename_axis("simulation_id", inplace=True)
        system_quantities_df.to_parquet(output_directory / "system_quantities.pq")


class VersionChoices(str, Enum):
    major = "major"
    minor = "minor"
    patch = "patch"
    post = "post"


@dev_app.command()
def publish_release(
    version_to_bump: Annotated[
        VersionChoices, typer.Argument(help="The version to bump", case_sensitive=False)
    ],
    version: Annotated[
        str,
        typer.Option(
            "--version",
            "-v",
            help="The version number to be published. Example: '1.1.2'",
        ),
    ] = None,
    dry_run: Annotated[
        bool,
        typer.Option(
            "--dry-run",
            "-d",
            is_flag=True,
            help="If set, the release will not be published, only the command to do so will be printed.",
        ),
    ] = False,
):
    """
    Publish a new RidePy release on GitHub.

    Requirements

    - The current branch must be `master`.
    - There must be no uncommitted changes.
    - The version in `pyproject.toml` must match the latest git tag.
    - The new version must be greater than the current one.
    - The `gh` command must be available and authenticated.
    - Git push access to the upstream ridepy repository must be available using the SSH agent credentials.
    - A private PGP key must be available to sign the commit.

    """
    working_dir = os.getcwd()
    repository_path = Path(pygit2.discover_repository(working_dir)).parent
    repo = pygit2.Repository(repository_path)

    if not repo.head.shorthand == "master":
        raise ValueError("Not on master branch, aborting.")

    if repo.status():
        raise ValueError("Uncommitted changes, aborting.")

    repo.remotes["upstream"].fetch(
        callbacks=pygit2.RemoteCallbacks(credentials=pygit2.KeypairFromAgent("git")),
    )
    remote_master_id = repo.lookup_reference("refs/remotes/upstream/master").target
    merge_result, _ = repo.merge_analysis(remote_master_id)

    if not merge_result == GIT_MERGE_ANALYSIS_UP_TO_DATE:
        raise ValueError(
            "Local master branch is not up-to-date with upstream, aborting."
        )

    pyproject_path = repository_path / "pyproject.toml"
    print(f"Discovered pyproject.toml at {pyproject_path}")

    with pyproject_path.open("rb") as fp:
        pyproject = tomli.load(fp)

    current_version_pyproject = packaging.version.parse(pyproject["project"]["version"])
    regex = re.compile(r"^refs/tags/v(.+)$")
    git_versions = []
    for r in repo.references:
        match = regex.match(r)
        if match is not None:
            git_versions.append(packaging.version.parse(match.groups()[0]))

    current_version_git = sorted(git_versions)[-1]

    if current_version_git != current_version_pyproject:
        raise ValueError(
            f"Version mismatch between pyproject.toml and git tags: "
            f"{current_version_pyproject} != {current_version_git}"
        )

    if version is None:
        if version_to_bump == "major":
            version = f"{current_version_git.major + 1}.0"
        if version_to_bump == "minor":
            version = f"{current_version_git.major}.{current_version_git.minor + 1}"
        if version_to_bump == "patch":
            version = (
                f"{current_version_git.major}.{current_version_git.minor}"
                f".{current_version_git.micro + 1}"
            )
        if version_to_bump == "post":
            version = (
                f"{current_version_git.major}.{current_version_git.minor}"
                f".{current_version_git.micro}"
                f".post{(current_version_git.post or 0) + 1}"
            )

    version = packaging.version.parse(version)

    print(f"New version is {version}")

    if current_version_git >= version:
        raise ValueError(
            f"New version ({version}) must be greater then "
            f"the current one ({current_version_git})."
        )
    else:
        print(f"Determined current version at {current_version_git}, continuing...")

    version = str(version)

    pyproject["project"]["version"] = version

    if not dry_run:
        with pyproject_path.open("wb") as fp:
            tomli_w.dump(pyproject, fp, multiline_strings=True)
            print("Updated pyproject.toml")
    else:
        print("Would update pyproject.toml, instead here comes the printout:\n")
        print(tomli_w.dumps(pyproject, multiline_strings=True))

    assert repo.status() == {
        "pyproject.toml": GIT_STATUS_WT_MODIFIED
    }, "pyproject.toml not modified. Aborting."

    repo.index.add("pyproject.toml")
    repo.index.write()

    assert repo.status() == {
        "pyproject.toml": GIT_STATUS_INDEX_MODIFIED
    }, "pyproject.toml not staged. Aborting."

    commit_string = repo.create_commit_string(
        repo.default_signature,  # author
        repo.default_signature,  # committer
        f"ridepy {version}",  # message
        repo.index.write_tree(),  # tree
        [repo.head.target],  # parents
    )

    gpg = gnupg.GPG()
    signed_commit = gpg.sign(commit_string, detach=True)
    commit = repo.create_commit_with_signature(
        commit_string, signed_commit.data.decode("utf-8")
    )
    repo.head.set_target(commit)

    print("Created signed commit.")

    assert not repo.status(), "Uncommitted changes after commit. Aborting."

    repo.remotes["upstream"].push(
        ["refs/heads/master"],
        callbacks=pygit2.RemoteCallbacks(credentials=pygit2.KeypairFromAgent("git")),
    )

    print("Pushed to upstream/master.")

    gh_cmd = [
        "gh",
        "release",
        "create",
        f"v{version}",
        "--generate-notes",
        "-t",
        f"ridepy {version}",
    ]
    if not dry_run:
        subprocess.run(gh_cmd)
    else:
        print("Would execute: " + " ".join(gh_cmd))

    print("Created release tag and GitHub release.")
    print("Done.")


@app.callback()
def main():
    """ridepy command-line interface"""

    pass


if __name__ == "__main__":
    app()



# distutils: language = c++

from libcpp.vector cimport vector
from libcpp.pair cimport pair
from libcpp.memory cimport shared_ptr

cdef extern from "cdata_structures.h" namespace 'ridepy':

    ctypedef pair[double, double] R2loc
    cpdef enum class StopAction(int):
        pickup=1
        dropoff=2
        internal=3

    cdef cppclass Request[Loc]:
        int request_id
        double creation_timestamp

    cdef cppclass TransportationRequest[Loc](Request[Loc]):
        Loc origin
        Loc destination
        double pickup_timewindow_min
        double pickup_timewindow_max
        double delivery_timewindow_min
        double delivery_timewindow_max

        TransportationRequest()
        TransportationRequest(int, double, Loc, Loc, double, double, double, double)

    cdef cppclass InternalRequest[Loc](Request[Loc]):
        Loc location

        InternalRequest()
        InternalRequest(int, double, Loc)


    cdef cppclass Stop[Loc]:
        Loc location
        shared_ptr[Request[Loc]] request
        StopAction action
        double estimated_arrival_time
        double time_window_min
        double time_window_max
        int occupancy_after_servicing

        Stop()
        Stop(Loc, const shared_ptr[Request]&, StopAction, double, int, double, double)

    cdef cppclass InsertionResult[Loc]:
        vector[Stop[Loc]] new_stoplist
        double min_cost
        double EAST_pu
        double LAST_pu
        double EAST_do
        double LAST_do

    cdef cppclass SingleVehicleSolution:
        int vehicle_id
        double min_cost
        double EAST_pu
        double LAST_pu
        double EAST_do
        double LAST_do




# distutils: language=c++

from libcpp.vector cimport vector
from libcpp.string cimport string
from libcpp.memory cimport shared_ptr
from libcpp.utility cimport pair
from ridepy.data_structures_cython.cdata_structures cimport (
     SingleVehicleSolution, TransportationRequest, Stop)
from ridepy.data_structures_cython.data_structures cimport StopAction
from ridepy.util.spaces_cython.cspaces cimport TransportSpace

cdef extern from "cvehicle_state.h" namespace 'ridepy':
    cdef cppclass StopEventSpec:
        StopAction action
        int request_id
        int vehicle_id
        double timestamp

    cdef cppclass VehicleState[Loc]:
        int vehicle_id
        vector[Stop[Loc]] stoplist
        int seat_capacity
        string dispatcher
        TransportSpace[Loc] &space

        VehicleState(
                int vehicle_id,
                vector[Stop[Loc]] initial_stoplist,
                TransportSpace[Loc] &space,
                string dispatcher,
                int seat_capacity
        )

        vector[StopEventSpec] fast_forward_time(double t)

        SingleVehicleSolution handle_transportation_request_single_vehicle(
                shared_ptr[TransportationRequest[Loc]] request
        )

        void select_new_stoplist()


# distutils: language=c++
import warnings

from ridepy.util import MAX_SEAT_CAPACITY

from ridepy.events import PickupEvent, DeliveryEvent, InternalEvent
from ridepy.data_structures import  (TransportationRequest as pyTransportationRequest,SingleVehicleSolution as pySingleVehicleSolution)
from ridepy.data_structures_cython.data_structures cimport (
    TransportationRequest,
    Stop,
    StopAction,
    Stoplist,
    LocType
)

from ridepy.data_structures_cython.cdata_structures cimport R2loc, Stop as CStop
from ridepy.data_structures_cython.cdata_structures cimport SingleVehicleSolution, \
    TransportationRequest as CTransportationRequest, \
    Request as CRequest

from ridepy.util.spaces_cython.spaces cimport TransportSpace
from ridepy.util.dispatchers_cython.dispatchers cimport Dispatcher

from typing import List, Union
from copy import deepcopy

import logging
logger = logging.getLogger(__name__)

from ridepy.vehicle_state_cython.cvehicle_state cimport VehicleState as CVehicleState, StopEventSpec
from libcpp.memory cimport unique_ptr, make_unique
from libcpp.utility cimport pair
from libcpp.memory cimport make_shared
from libcpp.vector cimport vector
from libcpp.string cimport string
from cython.operator cimport dereference

cdef extern from "limits.h":
    cdef int INT_MAX


cdef union _UVehicleState:
    unique_ptr[CVehicleState[R2loc]] _vstate_r2loc
    unique_ptr[CVehicleState[int]] _vstate_int

cdef class VehicleState:
    """
    Single vehicle insertion logic is implemented in Cython here. Can be used interchangeably
    with its pure-python equivalent `.vehicle_state.VehicleState`.
    """
    cdef Stoplist initial_stoplist
    cdef _UVehicleState _uvstate
    cdef LocType loc_type

    def __init__(
        self,
        #*, # haven't figured out yet how to get __reduce__+unpickling to work with keyword-only
        # arguments, hence we have to support __init__ with positional arguments for the time being
        int vehicle_id,
        initial_stoplist,
        TransportSpace space,
        Dispatcher dispatcher,
        int seat_capacity,
    ):
        # Create a cython stoplist object from initial_stoplist
        if isinstance(initial_stoplist, Stoplist):
            # if a `data_structures_cython.Stoplist` object, no need to re-create the cythonic stoplist
            self.initial_stoplist = initial_stoplist
        else:
            # assume that a python list of `data_structures_cython.Stop` objects are being passed
            # create a `data_structures_cython.Stoplist` object
            self.initial_stoplist = Stoplist(initial_stoplist, space.loc_type)
        if seat_capacity > INT_MAX:
            raise ValueError("Cannot use seat_capacity bigger that C++'s INT_MAX")
        if space.loc_type == LocType.R2LOC:
            self.loc_type = LocType.R2LOC
            self._uvstate._vstate_r2loc = make_unique[CVehicleState[R2loc]](
                vehicle_id, self.initial_stoplist.ustoplist._stoplist_r2loc,
                dereference(space.u_space.space_r2loc_ptr), dereference(dispatcher.u_dispatcher.dispatcher_r2loc_ptr), seat_capacity)
        elif space.loc_type == LocType.INT:
            self.loc_type = LocType.INT
            self._uvstate._vstate_int = make_unique[CVehicleState[int]](
                vehicle_id, self.initial_stoplist.ustoplist._stoplist_int,
                dereference(space.u_space.space_int_ptr), dereference(dispatcher.u_dispatcher.dispatcher_int_ptr), seat_capacity)
        else:
            raise ValueError("This line should never have been reached")



    def fast_forward_time(self, t: float) -> Tuple[List[StopEvent], List[Stop]]:
        """
        Update the vehicle_state to the simulator time `t`.

        Parameters
        ----------
        t
            time to be updated to

        Returns
        -------
        events
            List of stop events emitted through servicing stops upto time=t
        new_stoplist
            Stoplist remaining after servicing the stops upto time=t
        """
        # TODO assert that the CPATs are updated and the stops sorted accordingly
        # TODO optionally validate the travel time velocity constraints

        cdef vector[StopEventSpec] res
        if self.loc_type == LocType.R2LOC:
            res = dereference(self._uvstate._vstate_r2loc).fast_forward_time(t)
        elif self.loc_type == LocType.INT:
            res = dereference(self._uvstate._vstate_int).fast_forward_time(t)
        else:
            raise ValueError("This line should never have been reached")

        #stop_event_specc, stoplist = res[0],
        event_cache = []
        cdef StopEventSpec evspec

        for ev in res:
            if ev.action == StopAction.pickup:
                stop_event = {
                    "event_type": "PickupEvent",
                    "timestamp": ev.timestamp,
                    "request_id": ev.request_id,
                    "vehicle_id": ev.vehicle_id,
                }
            elif ev.action == StopAction.dropoff:
                stop_event = {
                    "event_type": "DeliveryEvent",
                    "timestamp": ev.timestamp,
                    "request_id": ev.request_id,
                    "vehicle_id": ev.vehicle_id,
                }
            elif ev.action == StopAction.internal:
                stop_event = {
                    "event_type": "InternalEvent",
                    "timestamp": ev.timestamp,
                    "vehicle_id": ev.vehicle_id,
                }
            else:
                raise ValueError(f"Unknown StopAction={ev.action}")

            event_cache.append(stop_event)

        return event_cache

    def handle_transportation_request_single_vehicle(
            self, request: pyTransportationRequest
    ) -> pySingleVehicleSolution:
        """
        The computational bottleneck. An efficient simulator could:

        1. Parallelize this over all vehicles. This function being without any side effects, it should be easy to do.

        Parameters
        ----------
        request
            Request to be handled.

        Returns
        -------
            The `SingleVehicleSolution` for the respective vehicle.
        """
        cdef SingleVehicleSolution single_vehicle_solution

        if self.loc_type == LocType.R2LOC:
            single_vehicle_solution = (
                dereference(self._uvstate._vstate_r2loc).handle_transportation_request_single_vehicle(
                    make_shared[CTransportationRequest[R2loc]](
                        <int> request.request_id,
                        <double> request.creation_timestamp,
                        <R2loc> request.origin,
                        <R2loc> request.destination,
                        <double> request.pickup_timewindow_min,
                        <double> request.pickup_timewindow_max,
                        <double> request.delivery_timewindow_min,
                        <double> request.delivery_timewindow_max
                    )
                )
            )
        elif self.loc_type == LocType.INT:
            single_vehicle_solution = (
                dereference(self._uvstate._vstate_int).handle_transportation_request_single_vehicle(
                    make_shared[CTransportationRequest[int]](
                        <int> request.request_id,
                        <double> request.creation_timestamp,
                        <int> request.origin,
                        <int> request.destination,
                        <double> request.pickup_timewindow_min,
                        <double> request.pickup_timewindow_max,
                        <double> request.delivery_timewindow_min,
                        <double> request.delivery_timewindow_max
                    )
                )
            )
        else:
            raise ValueError("This line should never have been reached")


        return (
            self.vehicle_id,
            single_vehicle_solution.min_cost,
            (
                single_vehicle_solution.EAST_pu,
                single_vehicle_solution.LAST_pu,
                single_vehicle_solution.EAST_do,
                single_vehicle_solution.LAST_do
            )
        )

    def select_new_stoplist(self):
        if self.loc_type == LocType.R2LOC:
            return dereference(self._uvstate._vstate_r2loc).select_new_stoplist()
        elif self.loc_type == LocType.INT:
            return dereference(self._uvstate._vstate_int).select_new_stoplist()
        else:
            raise ValueError("This line should never have been reached")

    property stoplist:
        def __get__(self):
            warnings.warn("Extracting the C++ stoplist will impact performance", UserWarning)
            if self.loc_type == LocType.R2LOC:
                return Stoplist.from_c_r2loc(dereference(self._uvstate._vstate_r2loc).stoplist)
            elif self.loc_type == LocType.INT:
                return Stoplist.from_c_int(dereference(self._uvstate._vstate_int).stoplist)
            else:
                raise ValueError("This line should never have been reached")

    property seat_capacity:
        def __get__(self):
            if self.loc_type == LocType.R2LOC:
                return dereference(self._uvstate._vstate_r2loc).seat_capacity
            elif self.loc_type == LocType.INT:
                return dereference(self._uvstate._vstate_int).seat_capacity
            else:
                raise ValueError("This line should never have been reached")

    property vehicle_id:
        def __get__(self):
            if self.loc_type == LocType.R2LOC:
                return dereference(self._uvstate._vstate_r2loc).vehicle_id
            elif self.loc_type == LocType.INT:
                return dereference(self._uvstate._vstate_int).vehicle_id
            else:
                raise ValueError("This line should never have been reached")



# distutils: language=c++

from cython.operator cimport dereference
from libcpp.memory cimport dynamic_pointer_cast
from ridepy.util.spaces_cython.spaces cimport TransportSpace

from ridepy.data_structures_cython.data_structures cimport (
    TransportationRequest,
    Stoplist,
    LocType,
    R2loc
)

from ridepy.data_structures_cython.cdata_structures cimport (
    InsertionResult,
    TransportationRequest as CTransportationRequest,
    Request as CRequest
)

from ridepy.util.dispatchers_cython.cdispatchers cimport (
    brute_force_total_traveltime_minimizing_dispatcher as c_brute_force_total_traveltime_minimizing_dispatcher,
    simple_ellipse_dispatcher as c_simple_ellipse_dispatcher
)


#################################################################################
# The dispatchers contained in this file are only used for testing purposes.
# They represent a wrapping of the C++ dispatchers through Cython to make them
# directly callable from python testing code. The Cython dispatchers now used for
# doing actual simulations don't make the C++ dispatchers callable from the python
# side anymore, they merely manage a pointer to a callable object which is only
# ever called from within the C++ vehicle state.
# At some point in the future these may be removed and the tests transitioned
# to pure C++/Cython.
#################################################################################

# Just like we did in data_structures_cython.Stop, we would have liked to have an union holding
# InsertionResult[R2loc] and InsertionResult[int] inside brute_force_total_traveltime_minimizing_dispatcher. However,
# that is nontrivial because any non-POD union member needs to have explicitly defined constructor and copy constructor
# (https://en.wikipedia.org/wiki/C%2B%2B11#Unrestricted_unions). Hence, the following union does *not* work, and we
# need to store two InsertionResult objects (one for each type) inside brute_force_total_traveltime_minimizing_dispatcher
#cdef union _UInsertionResult:
#     InsertionResult[R2loc] insertion_result_r2loc
#     InsertionResult[int] insertion_result_int



# This cpdef is crucial, otherwise we can't use this function from both python  and cython
cdef class BruteForceTotalTravelTimeMinimizingDispatcher:
    cdef LocType loc_type

    def __init__(self, loc_type):
        self.loc_type = loc_type

    def __call__(
            self,
            TransportationRequest cy_request,
            Stoplist stoplist,
            TransportSpace space,
            int seat_capacity,
            bint debug=False
    ):
        cdef InsertionResult[R2loc] insertion_result_r2loc
        cdef InsertionResult[int] insertion_result_int

        if self.loc_type == LocType.R2LOC:
            insertion_result_r2loc = c_brute_force_total_traveltime_minimizing_dispatcher[R2loc](
                dynamic_pointer_cast[CTransportationRequest[R2loc], CRequest[R2loc]](cy_request._ureq._req_r2loc),
                stoplist.ustoplist._stoplist_r2loc,
                dereference(space.u_space.space_r2loc_ptr), seat_capacity, debug
            )
            return insertion_result_r2loc.min_cost, Stoplist.from_c_r2loc(insertion_result_r2loc.new_stoplist), \
                   (insertion_result_r2loc.EAST_pu, insertion_result_r2loc.LAST_pu,
                    insertion_result_r2loc.EAST_do, insertion_result_r2loc.LAST_do)
        elif self.loc_type == LocType.INT:
            insertion_result_int = c_brute_force_total_traveltime_minimizing_dispatcher[int](
                dynamic_pointer_cast[CTransportationRequest[int], CRequest[int]](cy_request._ureq._req_int),
                stoplist.ustoplist._stoplist_int,
                dereference(space.u_space.space_int_ptr), seat_capacity, debug
            )
            return insertion_result_int.min_cost, Stoplist.from_c_int(insertion_result_int.new_stoplist), \
                   (insertion_result_int.EAST_pu, insertion_result_int.LAST_pu,
                    insertion_result_int.EAST_do, insertion_result_int.LAST_do)
        else:
            raise ValueError("This line should never have been reached")

cdef class SimpleEllipseDispatcher:
    cdef LocType loc_type

    def __init__(self, loc_type):
        self.loc_type = loc_type

    def __call__(
            self,
            TransportationRequest cy_request,
            Stoplist stoplist,
            TransportSpace space,
            int seat_capacity,
            double max_relative_detour=0,
            bint debug=False
    ):

        cdef InsertionResult[R2loc] insertion_result_r2loc
        cdef InsertionResult[int] insertion_result_int

        if self.loc_type == LocType.R2LOC:
            insertion_result_r2loc = c_simple_ellipse_dispatcher[R2loc](
                dynamic_pointer_cast[CTransportationRequest[R2loc], CRequest[R2loc]](cy_request._ureq._req_r2loc),
                stoplist.ustoplist._stoplist_r2loc,
                dereference(space.u_space.space_r2loc_ptr), seat_capacity, max_relative_detour, debug
            )
            return insertion_result_r2loc.min_cost, Stoplist.from_c_r2loc(insertion_result_r2loc.new_stoplist), \
                   (insertion_result_r2loc.EAST_pu, insertion_result_r2loc.LAST_pu,
                    insertion_result_r2loc.EAST_do, insertion_result_r2loc.LAST_do)
        elif self.loc_type == LocType.INT:
            insertion_result_int = c_simple_ellipse_dispatcher[int](
                dynamic_pointer_cast[CTransportationRequest[int], CRequest[int]](cy_request._ureq._req_int),
                stoplist.ustoplist._stoplist_int,
                dereference(space.u_space.space_int_ptr), seat_capacity, max_relative_detour, debug
            )
            return insertion_result_int.min_cost, Stoplist.from_c_int(insertion_result_int.new_stoplist), \
                   (insertion_result_int.EAST_pu, insertion_result_int.LAST_pu,
                    insertion_result_int.EAST_do, insertion_result_int.LAST_do)
        else:
            raise ValueError("This line should never have been reached")




from __future__ import annotations

import numpy as np

from typing import Optional, SupportsFloat, List, Tuple

from .data_structures import (
    Request,
    Stoplist,
    SingleVehicleSolution,
    StopAction,
    Stop,
    TransportationRequest,
    TransportSpace,
    Dispatcher,
    LocType,
)
from .events import PickupEvent, DeliveryEvent, InternalEvent, StopEvent

import logging

logger = logging.getLogger(__name__)


class VehicleState:
    """
    Single vehicle insertion logic is implemented here. Can optionally be implemented in Cython
    or another compiled language.
    """

    def recompute_arrival_times_drive_first(self):
        # update CPATs
        for stop_i, stop_j in zip(self.stoplist, self.stoplist[1:]):
            stop_j.estimated_arrival_time = max(
                stop_i.estimated_arrival_time, stop_i.time_window_min
            ) + self.space.t(stop_i.location, stop_j.location)

    def __init__(
        self,
        *,
        vehicle_id,
        initial_stoplist: Stoplist,
        space: TransportSpace,
        dispatcher: Dispatcher,
        seat_capacity: int,
    ):
        """
        Parameters
        ----------
        vehicle_id
            id of the vehicle to be created.
        initial_stoplist
            stoplist to start out with, MUST contain CPE as first element.
        space
        dispatcher
            see the docstring of `.FleetState`.
        seat_capacity
            the maximum number of `.TransportationRequest` s
             that can be in a vehicle at the same time.
        """
        self.vehicle_id = vehicle_id
        # TODO check for CPE existence in each supplied stoplist or encapsulate the whole thing
        self.stoplist: Stoplist = initial_stoplist
        """The list of `.Stop` objects specifying the planned future actions to be undertaken by this vehicle."""
        self.space = space
        self.dispatcher = dispatcher
        self.seat_capacity = seat_capacity

        logger.info(f"Created VehicleState with space of type {type(self.space)}")

    def fast_forward_time(self, t: float) -> Tuple[List[StopEvent], List[Stop]]:
        """
        Update the vehicle_state to the simulator time `t`.

        Parameters
        ----------
        t
            time to be updated to.

        Returns
        -------
        events
            List of stop events emitted through servicing stops upto time=t
        new_stoplist
            Stoplist remaining after servicing the stops upto time=t
        """
        # TODO assert that the CPATs are updated and the stops sorted accordingly
        # TODO optionally validate the travel time velocity constraints

        event_cache = []

        # Here, last_stop refers to the stop with the largest departure time value smaller or equal than t.
        # This can either be the last stop in the stoplist that is serviced here, or it can be the
        # (possibly outdated) CPE stop, of no other stop is serviced.
        last_stop = None

        # drop all non-future stops from the stoplist, except for the (outdated) CPE
        for i in range(len(self.stoplist) - 1, 0, -1):
            stop = self.stoplist[i]
            service_time = max(stop.estimated_arrival_time, stop.time_window_min)
            # service the stop at the minimum time at which it can leave
            if service_time <= t:
                # as we are iterating backwards, the first stop iterated over is the last one serviced
                if last_stop is None:
                    last_stop = stop

                if stop.action == StopAction.pickup:
                    stop_event = {
                        "event_type": "PickupEvent",
                        "timestamp": service_time,
                        "request_id": stop.request.request_id,
                        "vehicle_id": self.vehicle_id,
                    }
                elif stop.action == StopAction.dropoff:
                    stop_event = {
                        "event_type": "DeliveryEvent",
                        "timestamp": service_time,
                        "request_id": stop.request.request_id,
                        "vehicle_id": self.vehicle_id,
                    }
                elif stop.action == StopAction.internal:
                    stop_event = {
                        "event_type": "InternalEvent",
                        "timestamp": service_time,
                        "vehicle_id": self.vehicle_id,
                    }
                else:
                    raise ValueError(f"Unknown StopAction={stop.action}")

                event_cache.append(stop_event)

                del self.stoplist[i]

        # fix event cache order
        event_cache = event_cache[::-1]

        # if no stop was serviced, the last stop is the outdated CPE
        if last_stop is None:
            last_stop = self.stoplist[0]

        # set the occupancy at CPE
        self.stoplist[0].occupancy_after_servicing = last_stop.occupancy_after_servicing

        # set CPE location to current location as inferred from the time delta to the
        # upcoming stop's CPAT still mid-jump from last interpolation, no need to
        # interpolate again
        if self.stoplist[0].estimated_arrival_time <= t:
            if len(self.stoplist) > 1:
                loc, jump_time = self.space.interp_time(
                    u=last_stop.location,
                    v=self.stoplist[1].location,
                    time_to_dest=self.stoplist[1].estimated_arrival_time - t,
                )
                self.stoplist[0].location = loc
                # set CPE time
                self.stoplist[0].estimated_arrival_time = t + jump_time
            else:
                # Stoplist is (now) empty, only CPE is there. Set CPE time to
                # current time and move CPE to last_stop's location (which is
                # identical to CPE, if we haven't served anything.
                self.stoplist[0].location = last_stop.location
                self.stoplist[0].estimated_arrival_time = t

        return event_cache

    def handle_transportation_request_single_vehicle(
        self, request: TransportationRequest
    ) -> SingleVehicleSolution:
        """
        The computational bottleneck. An efficient simulator could do the following:

        1. Parallelize this over all vehicles. This function being without any side effects, it should be easy to do.
        2. Implement as a c extension. The args and the return value are all basic c data types, so this should also be easy.

        Parameters
        ----------
        request
            Request to be handled.

        Returns
        -------
        The `SingleVehicleSolution` for the respective vehicle.
        """

        cost, self.stoplist_new, (EAST_pu, LAST_pu, EAST_do, LAST_do) = self.dispatcher(
            request=request,
            stoplist=self.stoplist,
            space=self.space,
            seat_capacity=self.seat_capacity,
        )
        return self.vehicle_id, cost, (EAST_pu, LAST_pu, EAST_do, LAST_do)

    def select_new_stoplist(self):
        self.stoplist = self.stoplist_new



from copy import deepcopy

import numpy as np

from ridepy.data_structures import (
    TransportationRequest,
    Stoplist,
    TransportSpace,
    DispatcherSolution,
    Stop,
    StopAction,
)
from ridepy.util.dispatchers.dispatcher_class import dispatcherclass

import logging

logger = logging.getLogger(__name__)


def insert_request_to_stoplist_drive_first(
    stoplist: Stoplist,
    request: TransportationRequest,
    pickup_idx: int,
    dropoff_idx: int,
    space: TransportSpace,
    n_passengers: int = 1,  # future proofing when we have requests with multiple passengers
) -> Stoplist:
    """
    Inserts a request into  a stoplist. The pickup (dropoff) is inserted after pickup_idx (dropoff_idx).
    The estimated arrival times at all the stops are updated assuming a drive-first strategy.
    """
    # We don't want to modify stoplist in place. Make a copy.
    new_stoplist = deepcopy(stoplist)

    # Handle the pickup
    stop_before_pickup = new_stoplist[pickup_idx]
    cpat_at_pu = stop_before_pickup.estimated_departure_time + space.t(
        stop_before_pickup.location, request.origin
    )
    pickup_stop = Stop(
        location=request.origin,
        action=StopAction.pickup,
        estimated_arrival_time=cpat_at_pu,
        time_window_min=request.pickup_timewindow_min,
        time_window_max=request.pickup_timewindow_max,
        request=request,
        occupancy_after_servicing=stop_before_pickup.occupancy_after_servicing
        + n_passengers,
    )
    # increase the occupancies of all the stops between pickup and dropoff
    # remember, the indices are as follows:
    # 0,1,...,pickup_idx,(pickup_not_yet_inserted),...,dropoff_idx,(dropoff_not_yet_inserted), ...
    for s in new_stoplist[pickup_idx + 1 : dropoff_idx + 1]:
        s.occupancy_after_servicing += n_passengers

    insert_stop_to_stoplist_drive_first(new_stoplist, pickup_stop, pickup_idx, space)

    # Handle the dropoff
    dropoff_idx += 1
    stop_before_dropoff = new_stoplist[dropoff_idx]
    cpat_at_do = stop_before_dropoff.estimated_departure_time + space.t(
        stop_before_dropoff.location, request.destination
    )
    dropoff_stop = Stop(
        location=request.destination,
        action=StopAction.dropoff,
        estimated_arrival_time=cpat_at_do,
        time_window_min=request.delivery_timewindow_min,
        time_window_max=request.delivery_timewindow_max,
        request=request,
        occupancy_after_servicing=stop_before_dropoff.occupancy_after_servicing
        - n_passengers,
    )
    insert_stop_to_stoplist_drive_first(new_stoplist, dropoff_stop, dropoff_idx, space)

    return new_stoplist


def insert_stop_to_stoplist_drive_first(
    stoplist: Stoplist,
    stop: Stop,
    idx: int,
    space: TransportSpace,
) -> None:
    """
    Note: Modifies stoplist in-place. The passed stop has estimated_arrival_time set to None
    Args:
        stoplist:
        stop:
        idx:
        space:

    Returns:
    """
    stop_before_insertion = stoplist[idx]
    stop.estimated_arrival_time = cpat_of_inserted_stop(
        stop_before=stop_before_insertion,
        time_from_stop_before=space.t(stop_before_insertion.location, stop.location),
    )

    if idx < len(stoplist) - 1:
        # update CPATs of later stops
        delta_CPAT_next_stop = (
            stop.estimated_departure_time
            + space.t(stop.location, stoplist[idx + 1].location)
            - stoplist[idx + 1].estimated_arrival_time
        )

        for later_stop in stoplist[idx + 1 :]:
            old_departure = later_stop.estimated_departure_time
            later_stop.estimated_arrival_time += delta_CPAT_next_stop
            delta_CPAT_next_stop = later_stop.estimated_departure_time - old_departure

            if delta_CPAT_next_stop == 0:
                break

    stoplist.insert(idx + 1, stop)


def cpat_of_inserted_stop(
    stop_before: Stop, time_from_stop_before: float, delta_cpat: float = 0
) -> float:
    """
    Computes the cpat of the inserted stop, assuming drive first strategy.
    """
    return (
        max(
            stop_before.estimated_arrival_time + delta_cpat,
            stop_before.time_window_min,
        )
        + time_from_stop_before
    )


def time_to_stop_after_insertion(
    stoplist: Stoplist, location, index: int, space: TransportSpace
) -> float:
    """
    If a stop with `location` will have been inserted at `index`, computes the time
    from `location` to the stop after the insertion.

    Note: If the insertion is at the end of the stoplist, returns 0. Insertion at idx means after the idx'th stop.
    """

    return (
        space.t(location, stoplist[index + 1].location)
        if index < len(stoplist) - 1
        else 0
    )


def time_from_current_stop_to_next(
    stoplist: Stoplist, i: int, space: TransportSpace
) -> float:
    """
    Returns the time from the i'th stop in `stoplist` to the next stop.

    Note: If the insertion is at the end of the stoplist, returns 0
    """
    return (
        space.t(stoplist[i].location, stoplist[i + 1].location)
        if i < len(stoplist) - 1
        else 0
    )


def is_timewindow_violated_or_violation_worsened_due_to_insertion(
    stoplist: Stoplist, idx: int, est_arrival_first_stop_after_insertion: float
) -> bool:
    """
    If a stop is inserted at idx, so that the estimated_arrival_time at the stop after the inserted stop is
    est_arrival_first_stop_after_insertion, then checks for time window violations in the stoplist.

    Note: Assumes drive first strategy. Insertion at idx means after the idx'th stop.
    """

    # we are inserting at the end of the stoplist, nothing to check
    if idx > len(stoplist) - 2:
        return False

    # inserted stop incurs zero detour, and we don't have to wait
    if (
        est_arrival_first_stop_after_insertion
        <= stoplist[idx + 1].estimated_arrival_time
    ):
        return False

    delta_cpat = (
        est_arrival_first_stop_after_insertion
        - stoplist[idx + 1].estimated_arrival_time
    )

    for stop in stoplist[idx + 1 :]:
        old_leeway = stop.time_window_max - stop.estimated_arrival_time
        new_leeway = old_leeway - delta_cpat

        if new_leeway < 0 and new_leeway < old_leeway:
            return True
        elif stop.time_window_min >= stop.estimated_arrival_time + delta_cpat:
            # We have to wait or arrive just on time, thus no need to check next stops
            return False
        else:
            # Otherwise we are incurring additional delay. Compute the remaining delay:
            delta_cpat = (
                max(stop.time_window_min, stop.estimated_arrival_time + delta_cpat)
                - stop.estimated_departure_time
            )

    return False


@dispatcherclass
def BruteForceTotalTravelTimeMinimizingDispatcher(
    request: TransportationRequest,
    stoplist: Stoplist,
    space: TransportSpace,
    seat_capacity: int,
) -> DispatcherSolution:
    """
    Dispatcher that maps a vehicle's stoplist and a request to a new stoplist
    by minimizing the total driving time.

    This pure function is turned into a callable class by the decorator `.dispatcherclass` whose __init__ accepts
    optional arguments, see the docstring of `.dispatcherclass` for details.

    See the dispatcher interface in :ref:`desc_dispatcher` for details.


    Parameters
    ----------
    request
        request to be serviced.
    stoplist
        stoplist of the vehicle, to be mapped to a new stoplist.
    space
        transport space the vehicle is operating on.
    seat_capacity
            the maximum number of `.TransportationRequest` s that can be in a vehicle at the same time.

    Returns
    -------
        The best solution as defined in `.SingleVehicleSolution`.
    """
    min_cost = np.inf
    best_insertion = None
    for i, stop_before_pickup in enumerate(stoplist):
        if stop_before_pickup.occupancy_after_servicing == seat_capacity:
            # inserting here will violate capacity constraint
            continue
        time_to_pickup = space.t(stop_before_pickup.location, request.origin)
        CPAT_pu = cpat_of_inserted_stop(stop_before_pickup, time_to_pickup)
        # check for request's pickup timewindow violation
        if CPAT_pu > request.pickup_timewindow_max:
            continue
        EAST_pu = request.pickup_timewindow_min

        ######################
        # ADJACENT INSERTION #
        ######################
        CPAT_do = max(EAST_pu, CPAT_pu) + space.t(request.origin, request.destination)
        # check for request's dropoff timewindow violation
        if CPAT_do > request.delivery_timewindow_max:
            continue

        # compute the cost function
        time_to_dropoff = space.t(request.origin, request.destination)
        time_from_dropoff = time_to_stop_after_insertion(
            stoplist, request.destination, i, space
        )

        original_pickup_edge_length = time_from_current_stop_to_next(stoplist, i, space)
        total_cost = (
            time_to_pickup
            + time_to_dropoff
            + time_from_dropoff
            - original_pickup_edge_length
        )
        if total_cost < min_cost:
            # check for constraint violations at later points
            cpat_at_next_stop = (
                max(CPAT_do, request.delivery_timewindow_min) + time_from_dropoff
            )
            if not is_timewindow_violated_or_violation_worsened_due_to_insertion(
                stoplist, i, cpat_at_next_stop
            ):
                best_insertion = i, i
                min_cost = total_cost

        ##########################
        # NON-ADJACENT INSERTION #
        ##########################
        time_from_pickup = time_to_stop_after_insertion(
            stoplist, request.origin, i, space
        )
        cpat_at_next_stop = (
            max(CPAT_pu, request.pickup_timewindow_min) + time_from_pickup
        )
        if is_timewindow_violated_or_violation_worsened_due_to_insertion(
            stoplist, i, cpat_at_next_stop
        ):
            continue

        pickup_cost = time_to_pickup + time_from_pickup - original_pickup_edge_length

        if i < len(stoplist) - 1:
            delta_cpat = cpat_at_next_stop - stoplist[i + 1].estimated_arrival_time

        for j, stop_before_dropoff in enumerate(stoplist[i + 1 :], start=i + 1):
            # Need to check for seat capacity constraints. Note the loop: the constraint was not violated after
            # servicing the previous stop (otherwise we wouldn't've reached this line). Need to check that the
            # constraint is not violated due to the action at this stop (stop_before_dropoff)
            if stop_before_dropoff.occupancy_after_servicing == seat_capacity:
                # Capacity is violated. We need to break off this loop because no insertion either here or at a later
                # stop is permitted
                break
            time_to_dropoff = space.t(stop_before_dropoff.location, request.destination)
            CPAT_do = cpat_of_inserted_stop(
                stop_before_dropoff,
                time_to_dropoff,
                delta_cpat=delta_cpat,
            )
            # check for request's dropoff timewindow violation
            if CPAT_do > request.delivery_timewindow_max:
                break

            time_from_dropoff = time_to_stop_after_insertion(
                stoplist, request.destination, j, space
            )
            original_dropoff_edge_length = time_from_current_stop_to_next(
                stoplist, j, space
            )
            dropoff_cost = (
                time_to_dropoff + time_from_dropoff - original_dropoff_edge_length
            )
            total_cost = pickup_cost + dropoff_cost

            if total_cost < min_cost:
                # cost has decreased. check for constraint violations at later stops
                cpat_at_next_stop = (
                    max(CPAT_do, request.delivery_timewindow_min) + time_from_dropoff
                )
                if not is_timewindow_violated_or_violation_worsened_due_to_insertion(
                    stoplist, j, cpat_at_next_stop
                ):
                    best_insertion = i, j
                    min_cost = total_cost

            # we will try inserting the dropoff at a later stop
            # the delta_cpat is important to compute correctly for the next stop, it may have changed if
            # we had any slack time at this one
            new_departure_time = max(
                stop_before_dropoff.estimated_arrival_time + delta_cpat,
                stop_before_dropoff.time_window_min,
            )
            delta_cpat = (
                new_departure_time - stop_before_dropoff.estimated_departure_time
            )

    if min_cost < np.inf:
        best_pickup_idx, best_dropoff_idx = best_insertion
        # if request.request_id == 2:
        # print(f"Py DEBUG: best insertion @ {best_insertion}")
        # print(stoplist[0].estimated_arrival_time)
        # print(stoplist[0].location)
        # print(request.creation_timestamp)
        # print()

        logger.info(f"Best insertion: {best_insertion}")
        logger.info(f"Min cost: {min_cost}")

        new_stoplist = insert_request_to_stoplist_drive_first(
            stoplist=stoplist,
            request=request,
            pickup_idx=best_pickup_idx,
            dropoff_idx=best_dropoff_idx,
            space=space,
        )
        EAST_pu, LAST_pu = (
            new_stoplist[best_pickup_idx + 1].time_window_min,
            new_stoplist[best_pickup_idx + 1].time_window_max,
        )
        EAST_do, LAST_do = (
            new_stoplist[best_dropoff_idx + 2].time_window_min,
            new_stoplist[best_dropoff_idx + 2].time_window_max,
        )
        return min_cost, new_stoplist, (EAST_pu, LAST_pu, EAST_do, LAST_do)
    else:
        return min_cost, None, (np.nan, np.nan, np.nan, np.nan)



# distutils: language = c++

from libcpp.vector cimport vector
from libcpp.memory cimport shared_ptr


from ridepy.data_structures_cython.cdata_structures cimport (
    Request as CRequest,
    TransportationRequest as CTransportationRequest,
    InternalRequest as CInternalRequest,
    Stop as CStop,
    R2loc
)

cdef extern from * namespace 'ridepy':
    cpdef enum class StopAction(int):
        pickup=1
        dropoff=2
        internal=3

cpdef enum class LocType:
    R2LOC  = 1
    INT = 2


cdef union _URequest:
    shared_ptr[CRequest[R2loc]] _req_r2loc
    shared_ptr[CRequest[int]] _req_int

cdef union _UTransportationRequest:
    shared_ptr[CTransportationRequest[R2loc]] _req_r2loc
    shared_ptr[CTransportationRequest[int]] _req_int

cdef union _UInternalRequest:
    shared_ptr[CInternalRequest[R2loc]] _req_r2loc
    shared_ptr[CInternalRequest[int]] _req_int

cdef union _UStop:
    # have to store a pointer otherwise setting attribute from python side does not work
    CStop[R2loc] *_stop_r2loc
    CStop[int] *_stop_int

cdef union _UStoplist:
    vector[CStop[R2loc]] _stoplist_r2loc
    vector[CStop[int]] _stoplist_int


cdef class Request:
    cdef _URequest _ureq
    cdef LocType loc_type

cdef class TransportationRequest(Request):
    cdef _UTransportationRequest _utranspreq
    @staticmethod
    cdef TransportationRequest from_c_r2loc(shared_ptr[CTransportationRequest[R2loc]] creq)
    @staticmethod
    cdef TransportationRequest from_c_int(shared_ptr[CTransportationRequest[int]] creq)

cdef class InternalRequest(Request):
    cdef _UInternalRequest _uinternreq
    @staticmethod
    cdef InternalRequest from_c_r2loc(shared_ptr[CInternalRequest[R2loc]] creq)
    @staticmethod
    cdef InternalRequest from_c_int(shared_ptr[CInternalRequest[int]] creq)

cdef class Stop:
    cdef bint ptr_owner
    cdef _UStop ustop
    cdef LocType loc_type
    @staticmethod
    cdef Stop from_c_r2loc(CStop[R2loc] *cstop)
    @staticmethod
    cdef Stop from_c_int(CStop[int] *cstop)


cdef class Stoplist:
    cdef LocType loc_type
    cdef _UStoplist ustoplist
    cdef Stop py_s
    @staticmethod
    cdef Stoplist from_c_r2loc(vector[CStop[R2loc]] cstoplist)
    @staticmethod
    cdef Stoplist from_c_int(vector[CStop[int]] cstoplist)



import dataclasses
import importlib
import collections.abc
import json

import operator as op
import functools as ft
import numpy as np

from typing import Iterable
from pathlib import Path

from ridepy.data_structures import TransportSpace, DistanceDistribution
from ridepy.util.spaces_cython import TransportSpace as CyTransportSpace


class ParamsJSONEncoder(json.JSONEncoder):
    """
    JSONEncoder to use when serializing a dictionary containing simulation parameters.
    This is able to serialize `RequestGenerator`, `TransportSpace` and dispatchers.

    Example
    -------

    .. code-block:: python

        json.dumps(params, cls=ParamsJSONEncoder)

    """

    def default(self, obj):
        # request generator cls?
        if isinstance(obj, type):
            return f"{obj.__module__}.{obj.__qualname__}"
        # TransportSpace?
        elif isinstance(obj, (TransportSpace, CyTransportSpace, DistanceDistribution)):
            # TODO in future, large networks might be saved in another file to be reused
            return {
                f"{obj.__class__.__module__}.{obj.__class__.__name__}": obj.asdict()
            }
        # dispatcher?
        elif callable(obj):
            return f"{obj.__module__}.{obj.__name__}"
        elif isinstance(obj, Path):
            return str(obj.expanduser().resolve())
        elif isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return json.JSONEncoder.default(self, obj)


class ParamsJSONDecoder(json.JSONDecoder):
    """
    JSONDecoder to use when deserializing a dictionary containing simulation parameters.
    This is able to deserialize `RequestGenerator`, `TransportSpace` and dispatchers.

    Example
    -------

    .. code-block:: python

       json.loads(params, cls=ParamsJSONDecoder)

    """

    def __init__(self, *args, **kwargs):
        super().__init__(object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, dct):
        if "coord_range" in dct:
            dct["coord_range"] = [(a, b) for a, b in dct["coord_range"]]
        else:
            if "initial_location" in dct and isinstance(dct["initial_location"], list):
                dct["initial_location"] = tuple(dct["initial_location"])

            if "initial_locations" in dct and dct["initial_locations"] is not None:
                dct["initial_locations"] = {
                    int(vehicle_id): location
                    for vehicle_id, location in dct["initial_locations"].items()
                }

            for cls_str in [
                "transportation_request_cls",
                "fleet_state_cls",
                "vehicle_state_cls",
                "request_generator_cls",
                "dispatcher_cls",
            ]:
                if cls_str in dct:
                    module, cls = dct[cls_str].rsplit(".", 1)
                    dct[cls_str] = getattr(importlib.import_module(module), cls)

            for obj_str in ["space", "distance_distribution"]:
                if obj_str in dct:
                    path, kwargs = next(iter(dct[obj_str].items()))
                    module, cls = path.rsplit(".", 1)
                    dct[obj_str] = getattr(importlib.import_module(module), cls)(
                        **kwargs
                    )

            if "data_dir" in dct:
                dct["data_dir"] = Path(dct["data_dir"])

        return dct


class EventsJSONEncoder(json.JSONEncoder):
    """
    JSONEncoder to use when serializing a list containing `Event`.

    Example
    -------

    .. code-block:: python

        json.dumps(events, cls=EventsJSONEncoder)

    """

    def default(self, obj):
        if dataclasses.is_dataclass(obj):
            return {"event_type": obj.__class__.__name__} | dataclasses.asdict(obj)
        else:
            return json.JSONEncoder.default(self, obj)


def sort_params(params: dict) -> dict:
    """
    Returns a copy of the two-level nested dict `params` which is sorted
    in both levels.

    Parameters
    ----------
    params
        Parameter dictionary, two levels of nesting

    Returns
    -------
    params
        Sorted params dict
    """
    params = dict(sorted(params.items(), key=op.itemgetter(0)))
    for outer_key, inner_dict in params.items():
        params[outer_key] = dict(sorted(inner_dict.items(), key=op.itemgetter(0)))
    return params


def create_params_json(*, params: dict, sort=True) -> str:
    """
    Create a dictionary containing simulation parameters to pretty JSON.
    Parameter dictionaries may contain anything that is supported
    by `.ParamsJSONEncoder` and `.ParamsJSONDecoder`, e.g. `RequestGenerator`,
    `TransportSpace`s and dispatchers. For additional detail, see :ref:`Executing Simulations`.

    Parameters
    ----------
    params
        dictionary containing the params to save
    sort
        if sort is True, sort the dict recursively to ensure consistent order.
    """
    if sort:
        params = sort_params(params)
    return json.dumps(params, indent=4, cls=ParamsJSONEncoder)


def save_params_json(*, param_path: Path, params: dict) -> None:
    """
    Save a dictionary containing simulation parameters to pretty JSON,
    overwriting existing. Parameter dictionaries may contain anything that is supported
    by `.ParamsJSONEncoder` and `.ParamsJSONDecoder`, e.g. `RequestGenerator`,
    `TransportSpace`s and dispatchers. For additional detail, see :ref:`Executing Simulations`.

    Parameters
    ----------
    param_path
        JSON output file path
    params
        dictionary containing the params to save
    """
    with open(str(param_path), "w") as f:
        f.write(create_params_json(params=params))


def read_params_json(param_path: Path) -> dict:
    """
    Read a dictionary containing simulation parameters from JSON.
    Parameter dictionaries may contain anything that is supported
    by `.ParamsJSONEncoder` and `.ParamsJSONDecoder`, e.g. `RequestGenerator`,
    `TransportSpace`s and dispatchers. For additional detail, see :ref:`Executing Simulations`.

    Parameters
    ----------
    param_path

    Returns
    -------
    parameter dictionary
    """

    return json.load(param_path.open("r"), cls=ParamsJSONDecoder)


def save_events_json(*, jsonl_path: Path, events: Iterable) -> None:
    """
    Save events iterable to a file according to JSONL specs, appending to existing.
    For additional detail, see :ref:`Executing Simulations`.

    Parameters
    ----------
    jsonl_path
        JSON Lines output file path
    events
        iterable containing the events to save
    """
    with jsonl_path.open("a", encoding="utf-8") as f:
        for event in events:
            print(json.dumps(event, cls=EventsJSONEncoder), file=f)


def read_events_json(jsonl_path: Path) -> list[dict]:
    """
    Read events from JSON lines file, where each line of the file contains a single event.
    For additional detail, see :ref:`Executing Simulations`.

    Parameters
    ----------
    jsonl_path
       JSON Lines input file path

    Returns
    -------
    List of dicts
    """
    with jsonl_path.open("r", encoding="utf-8") as f:
        return list(map(json.loads, f.readlines()))



import pandas as pd


def _create_transportation_requests_dataframe(
    *, evs: pd.DataFrame, stops, space
) -> pd.DataFrame:
    """
    Create DataFrame containing all requests.
    This includes the *submitted* requests, the *accepted* requests and the *rejected* requests.
    For the accepted requests, additional *serviced* and subsequently *inferred* data is given,
    under the respectively named level-0 column index *source*. Level 1 specifying the actual
    kind data of data is named *quantity*.

    The schema of the returned DataFrame is the following,
    where `request_id` is the index.

    .. code-block::

        Column                               Dtype
        ------                               -----
        request_id                            int64
        (submitted, timestamp)                float64
        (submitted, origin)                   float64
        (submitted, destination)              float64
        (submitted, pickup_timewindow_min)    float64
        (submitted, pickup_timewindow_max)    float64
        (submitted, delivery_timewindow_min)  float64
        (submitted, delivery_timewindow_max)  float64
        (submitted, direct_travel_distance)   float64
        (submitted, direct_travel_time)       float64
        (accepted, timestamp)                 float64
        (accepted, origin)                    float64
        (accepted, destination)               float64
        (accepted, pickup_timewindow_min)     float64
        (accepted, pickup_timewindow_max)     float64
        (accepted, delivery_timewindow_min)   float64
        (accepted, delivery_timewindow_max)   float64
        (rejected, timestamp)                 float64
        (inferred, relative_travel_time)      float64
        (inferred, travel_time)               float64
        (inferred, waiting_time)              float64
        (serviced, timestamp_dropoff)         float64
        (serviced, timestamp_pickup)          float64
        (serviced, vehicle_id)                float64

    Parameters
    ----------
    evs
        events DataFrame
    stops
        stoplists DataFrame
    space
        TransportSpace the simulations were performed on

    Returns
    -------
    requests DataFrame indexed by `request_id`
    """

    # first turn all request submission, acceptance and rejection events into a dataframe
    reqs = (
        evs[
            evs["event_type"].isin(
                [
                    "RequestSubmissionEvent",
                    "RequestAcceptanceEvent",
                    "RequestRejectionEvent",
                ]
            )
        ]  # select only request events
        .drop(["vehicle_id", "location"], axis=1)  # drop now empty columns
        .set_index(
            ["request_id", "event_type"]
        )  # set index to be request_id and event_type
        .unstack()  # unstack so that remaining index is request_id and column index is MultiIndex event_type as level_1
        .reorder_levels(
            [1, 0], axis=1
        )  # switch column index order to have event_type as level_0
        .sort_index(axis=1)  # sort so that columns are grouped by event_type
        .drop(
            [
                ("RequestRejectionEvent", "pickup_timewindow_min"),
                ("RequestRejectionEvent", "pickup_timewindow_max"),
                ("RequestRejectionEvent", "delivery_timewindow_min"),
                ("RequestRejectionEvent", "delivery_timewindow_max"),
                ("RequestRejectionEvent", "origin"),
                ("RequestRejectionEvent", "destination"),
            ],
            axis=1,
            errors="ignore",
        )  # drop columns that are empty (nan) for rejections (errors=ignore b/c rejections may not have happened)
        .rename(
            {
                "RequestAcceptanceEvent": "accepted",
                "RequestSubmissionEvent": "submitted",
                "RequestRejectionEvent": "rejected",
            },
            axis=1,
            level=0,
        )  # rename event_types to the "data source" indicators "accepted", "submitted" and "rejected"
    )
    reqs.columns.rename(["source", "quantity"], inplace=True)

    # Get properties of serviced requests from the stops dataframe:
    stops_tmp = stops.reset_index()[
        ["request_id", "vehicle_id", "timestamp", "delta_occupancy"]
    ].set_index("request_id")

    # - vehicle ID of the vehicle that serviced the request
    reqs[("serviced", "vehicle_id")] = stops_tmp[stops_tmp["delta_occupancy"] > 0][
        "vehicle_id"
    ]

    # - timestamp of the pickup stop
    reqs[("serviced", "timestamp_pickup")] = stops_tmp[
        stops_tmp["delta_occupancy"] > 0
    ]["timestamp"]

    # - timestamp of the dropoff stop
    reqs[("serviced", "timestamp_dropoff")] = stops_tmp[
        stops_tmp["delta_occupancy"] < 0
    ]["timestamp"]

    # - travel time
    reqs[("inferred", "travel_time")] = (
        reqs[("serviced", "timestamp_dropoff")] - reqs[("serviced", "timestamp_pickup")]
    )

    # If transportation as submitted were submitted, calculate more properties.
    # NOTE that these properties might equally well be computed using the
    # inferred requests, but in case of differences between the requests
    # the resulting change in behavior might not be intended. Therefore, so far
    # we only compute these quantities if transportation_requests are submitted.

    # - direct travel time
    # `to_list()` is necessary as for dimensionality > 1 the `pd.Series` will contain tuples
    # which will not be understood as a dimension by `np.shape(...)` which subsequently confuses smartVectorize
    # see https://github.com/PhysicsOfMobility/ridepy/issues/85
    reqs[("submitted", "direct_travel_time")] = space.t(
        reqs[("submitted", "origin")].to_list(),
        reqs[("submitted", "destination")].to_list(),
    )

    # - direct travel distance
    # again: `to_list()` is necessary as for dimensionality > 1 the `pd.Series` will contain tuples
    # which will not be understood as a dimension by `np.shape(...)` which subsequently  confuses smartVectorize
    # see https://github.com/PhysicsOfMobility/ridepy/issues/85
    reqs[("submitted", "direct_travel_distance")] = space.d(
        reqs[("submitted", "origin")].to_list(),
        reqs[("submitted", "destination")].to_list(),
    )

    # - waiting time
    reqs[("inferred", "waiting_time")] = (
        reqs[("serviced", "timestamp_pickup")]
        - reqs[("submitted", "pickup_timewindow_min")]
    )

    # - relative travel time
    reqs[("inferred", "relative_travel_time")] = (
        reqs[("inferred", "travel_time")] / reqs[("submitted", "direct_travel_time")]
    )

    # TODO possibly add more properties to compute HERE

    # sort columns alphabetically
    # FIXME this is open for debate
    reqs.sort_values(["source", "quantity"], axis=1, inplace=True)
    return reqs



def dispatcherclass(f):
    """
    Use this decorator to create a callable Dispatcher class from a pure function. Use this to conveniently turn a
    pure function mapping a stoplist and a request to an updated stoplist into a dispatcher usable with ridepy.

    In principle, using dispatcher objects allow for easy configuration of dispatcher behavior. After instantiation
    of an object ``dispatcher``, the original dispatcher function is available as ``dispatcher(...)``. Currently on
    initiating a dispatcher object, ``loc_type`` can be supplied for interface compatibility with Cython dispatchers.

    Use like

    .. code-block:: python

        @dispatcherclass
        def MyFancyDispatcher(
            request: TransportationRequest,
            stoplist: Stoplist,
            space: TransportSpace,
            seat_capacity: int,
        ) -> SingleVehicleSolution: ...

    """

    class DispatcherClass:
        __name__ = f.__name__
        __qualname__ = f.__qualname__
        __doc__ = f.__doc__
        __module__ = f.__module__

        def __init__(self, loc_type=None):
            self.loc_type = loc_type

        def __call__(self, *args, **kwargs):
            return f(*args, **kwargs)

    return DispatcherClass



from typing import Iterable, List

import pandas as pd

from ridepy.data_structures import TransportSpace
from ridepy.util.analytics.stops import (
    _create_stoplist_dataframe,
    _add_locations_to_stoplist_dataframe,
)
from ridepy.util.analytics.requests import _create_transportation_requests_dataframe


def _create_events_dataframe(events: Iterable[dict]) -> pd.DataFrame:
    """
    Create a DataFrame of all logged events with their properties at columns.

    The schema of the returned DataFrame is the following, where
    the index is an unnamed integer range index.

    .. code-block::

        Column                   Dtype
        ------                   -----
        request_id               int64
        timestamp                float64
        vehicle_id               float64
        event_type               object
        location                 Union[int, float64, Tuple[float64]]
        origin                   Union[int, float64, Tuple[float64]]
        destination              Union[int, float64, Tuple[float64]]
        pickup_timewindow_min    float64
        pickup_timewindow_max    float64
        delivery_timewindow_min  float64
        delivery_timewindow_max  float64

    Parameters
    ----------
    events

    Returns
    -------
    events DataFrame, indexed by integer range
    """
    return pd.DataFrame(events)


def get_stops_and_requests_from_events_dataframe(
    *, events_df: pd.DataFrame, space: TransportSpace
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Prepare stops and requests dataframes from an events dataframe.
    For details on the returned dataframes see doc on outside-facing `get_stops_and_requests`.

    Parameters
    ----------
    events_df
        DataFrame indexed
    space

    Returns
    -------
    stops
        dataframe indexed by `[vehicle_id, timestamp]` containing all stops
    requests
        dataframe indexed by `request_id` containing all requests
    """
    stops_df = _create_stoplist_dataframe(evs=events_df)
    requests_df = _create_transportation_requests_dataframe(
        evs=events_df, stops=stops_df, space=space
    )

    # try:
    stops_df = _add_locations_to_stoplist_dataframe(
        reqs=requests_df, stops=stops_df, space=space
    )
    # stops_df = _add_insertion_stats_to_stoplist_dataframe(
    #     reqs=requests_df, stops=stops_df, space=space
    # )
    # except KeyError:  # TODO document this

    # pass

    return stops_df, requests_df


def get_stops_and_requests(
    *, events: List[dict], space: TransportSpace
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Prepare two DataFrames, containing stops and requests.

    # NOTE: This assumes occupancy delta of +1/-1, i.e. only single-customer requests.
    #       If the simulator should allow for multi-customer requests in the future,
    #       this must be changed.
    #       See also [issue #45](https://github.com/PhysicsOfMobility/ridepy/issues/45)

    The `stops` DataFrame returned has the following schema:

    .. code-block::

        Column                                       Dtype
        ------                                       -----
        vehicle_id                                 float64
        stop_id                                      int64
        timestamp                                  float64
        delta_occupancy                            float64
        request_id                                   int64
        state_duration                             float64
        occupancy                                  float64
        location                                    object
        dist_to_next                               float64
        time_to_next                               float64
        timestamp_submitted                        float64
        insertion_index                            float64
        leg_1_dist_service_time                    float64
        leg_2_dist_service_time                    float64
        leg_direct_dist_service_time               float64
        detour_dist_service_time                   float64
        leg_1_dist_submission_time                 float64
        leg_2_dist_submission_time                 float64
        leg_direct_dist_submission_time            float64
        detour_dist_submission_time                float64
        stoplist_length_submission_time            float64
        stoplist_length_service_time               float64
        avg_segment_dist_submission_time           float64
        avg_segment_time_submission_time           float64
        avg_segment_dist_service_time              float64
        avg_segment_time_service_time              float64
        system_stoplist_length_submission_time     float64
        system_stoplist_length_service_time        float64
        avg_system_segment_dist_submission_time    float64
        avg_system_segment_time_submission_time    float64
        avg_system_segment_dist_service_time       float64
        avg_system_segment_time_service_time       float64
        relative_insertion_position                float6


    The `requests` DataFrame returned has the following schema:

    .. code-block::

        Column                               Dtype
        ------                               -----
        (submitted, timestamp)                float64
        (submitted, origin)                   Union[float64, int, Tuple[float64]]
        (submitted, destination)              Union[float64, int, Tuple[float64]]
        (submitted, pickup_timewindow_min)    float64
        (submitted, pickup_timewindow_max)    float64
        (submitted, delivery_timewindow_min)  float64
        (submitted, delivery_timewindow_max)  float64
        (submitted, direct_travel_distance)   float64
        (submitted, direct_travel_time)       float64
        (accepted, timestamp)                 float64
        (accepted, origin)                    Union[float64, int, Tuple[float64]]
        (accepted, destination)               Union[float64, int, Tuple[float64]]
        (accepted, pickup_timewindow_min)     float64
        (accepted, pickup_timewindow_max)     float64
        (accepted, delivery_timewindow_min)   float64
        (accepted, delivery_timewindow_max)   float64
        (rejected, timestamp)                 float64
        (inferred, relative_travel_time)      float64
        (inferred, travel_time)               float64
        (inferred, waiting_time)              float64
        (serviced, timestamp_dropoff)         float64
        (serviced, timestamp_pickup)          float64
        (serviced, vehicle_id)                float64

    Parameters
    ----------
    events
        list of all the events returned by the simulation
    space
        transportation space that was used for the simulations

    Returns
    -------
    stops
        dataframe indexed by `[vehicle_id, timestamp]` containing all stops
    requests
        dataframe indexed by `request_id` containing all requests
    """

    return get_stops_and_requests_from_events_dataframe(
        events_df=_create_events_dataframe(events=events), space=space
    )



from ridepy.util.dispatchers.taxicab import TaxicabDispatcherDriveFirst
from ridepy.util.dispatchers.ridepooling import (
    BruteForceTotalTravelTimeMinimizingDispatcher,
)

"""
This package contains pure python dispatchers. These are callable from python and should be used for testing 
purposes and small scale simulations where computing performance is not of primary concern. For larger scale 
simulations, use Cython dispatchers. Note that those can not be called from python directly, but through the 
`.vehicle_state_cython.VehicleState` class.
"""



# distutils: language=c++

from libcpp.vector cimport vector
from libcpp.memory cimport shared_ptr
from ridepy.data_structures_cython.cdata_structures cimport (
    InsertionResult, TransportationRequest, Stop)
from ridepy.util.spaces_cython.cspaces cimport TransportSpace, Euclidean2D


cdef extern from "cdispatchers.h" namespace 'ridepy':
    InsertionResult[Loc] brute_force_total_traveltime_minimizing_dispatcher[Loc](
          shared_ptr[TransportationRequest[Loc]] request,
          vector[Stop[Loc]] &stoplist,
          const TransportSpace &space, int seat_capacity, bint debug)

    InsertionResult[Loc] simple_ellipse_dispatcher[Loc](
          shared_ptr[TransportationRequest[Loc]] request,
          vector[Stop[Loc]] &stoplist,
          const TransportSpace &space,
          int seat_capacity,
          double max_relative_detour,
          bint debug
    )

    cdef cppclass AbstractDispatcher[Loc]:
        AbstractDispatcher()

    cdef cppclass BruteForceTotalTravelTimeMinimizingDispatcher[Loc](AbstractDispatcher[Loc]):
        BruteForceTotalTravelTimeMinimizingDispatcher()

    cdef cppclass SimpleEllipseDispatcher[Loc](AbstractDispatcher[Loc]):
        SimpleEllipseDispatcher(double)



from ridepy.util.analytics.events import (
    get_stops_and_requests,
    get_stops_and_requests_from_events_dataframe,
)
from ridepy.util.analytics.vehicle import get_vehicle_quantities
from ridepy.util.analytics.system import get_system_quantities



import pandas as pd

from typing import Optional, Any, Union


def get_system_quantities(
    stops: pd.DataFrame,
    requests: pd.DataFrame,
    params: Optional[dict[str, dict[str, Any]]] = None,
) -> dict[str, Union[int, float]]:
    """
    Compute various quantities aggregated for the entire simulation.

    Currently, the following observables are returned (quantities in parentheses may not
    be returned if ``params`` is not given/``stops`` does not contain the respective
    quantities):

    - **(n_vehicles)** -- number of vehicles (simulation parameter)
    - **(request_rate)** -- request rate (simulation parameter)
    - **(velocity)** -- vehicle velocity (simulation parameter)
    - **(load_requested)** -- load requested (derived from request rate, velocity, and number of vehicles as input parameters and the spaces' average distance)
    - **(load_serviced)** -- load requested (derived from request rate times (1-rejection ratio), velocity, and number of vehicles as input parameters and the spaces' average distance)
    - **(avg_direct_dist_space)** -- average direct distance in space, computed by taking 1e5 random samples
    - **avg_occupancy**
    - **avg_segment_dist**
    - **avg_segment_time**
    - **total_dist_driven**
    - **total_time_driven**
    - **avg_direct_dist**
    - **avg_direct_time**
    - **total_direct_dist**
    - **total_direct_time**
    - **efficiency_dist**
    - **efficiency_time**
    - **avg_waiting_time**
    - **rejection_ratio**
    - **median_stoplist_length** -- median per-vehicle stoplist length, taken over all "stoplist states" (vehicles x time)
    - **median_stoplist_length** -- median per-vehicle stoplist length, taken over all "stoplist states" (vehicles x time)
    - **mean_system_stoplist_length** -- arithmetic mean system-wide stoplist length, taken over all "stoplist states" (time)
    - **mean_system_stoplist_length** -- arithmetic mean system-wide stoplist length, taken over all "stoplist states" (time)
    - **avg_detour**
    - **(avg_system_stoplist_length_service_time)**
    - **(avg_system_stoplist_length_submission_time)**
    - **(avg_stoplist_length_service_time)**
    - **(avg_stoplist_length_submission_time)**

    Parameters
    ----------
    stops
        Stops dataframe
    requests
        Requests dataframe
    params
        Optional, adds more data (the fields in parentheses) to the result for convenience purposes

    Returns
    -------
    system_quantities
        dict containing the aforementioned observables

    """
    serviced_requests = (
        requests[requests[("rejected", "timestamp")].isna()]
        if ("rejected", "timestamp") in requests
        else requests
    )

    avg_occupancy = (stops["occupancy"] * stops["state_duration"]).sum() / stops[
        "state_duration"
    ].sum()

    avg_segment_dist = stops["dist_to_next"].mean()
    avg_segment_time = stops["time_to_next"].mean()

    total_dist_driven = stops["dist_to_next"].sum()
    total_time_driven = stops["time_to_next"].sum()

    avg_direct_dist = serviced_requests[("submitted", "direct_travel_distance")].mean()
    avg_direct_time = serviced_requests[("submitted", "direct_travel_time")].mean()

    total_direct_dist = serviced_requests[("submitted", "direct_travel_distance")].sum()
    total_direct_time = serviced_requests[("submitted", "direct_travel_time")].sum()

    efficiency_dist = total_direct_dist / total_dist_driven
    efficiency_time = total_direct_time / total_time_driven

    avg_waiting_time = serviced_requests.inferred.waiting_time.mean()

    rejection_ratio = 1 - len(serviced_requests) / len(requests)

    stops["event_type"] = stops["delta_occupancy"].map({1.0: "pickup", -1.0: "dropoff"})

    submission_events = requests.loc[
        :, [("submitted", "timestamp"), ("serviced", "vehicle_id")]
    ].dropna()
    submission_events.columns = ["timestamp", "vehicle_id"]
    submission_events = (
        submission_events.reset_index()
        .set_index(["vehicle_id", "timestamp"])
        .sort_index()
        .assign(event_type="submission")
    )

    event_log = pd.concat(
        [
            stops.reset_index().set_index(["vehicle_id", "timestamp"])[
                ["event_type", "request_id"]
            ],
            submission_events,
        ],
        axis="index",
    ).sort_index()

    # This computes the median stoplist length. The median is taken over all states
    # of all vehicles, all states weighted equally. State changes are incurred by
    # submission events, pickup stops, delivery stops and internal stops
    # (initial and final stops). State changes at the same timestamp are merged.
    # It assumes that the stoplist is ordered by timestamp.
    stoplist_length_ensemble = (
        event_log["event_type"]
        .map(dict(submission=2, pickup=-1, dropoff=-1))
        .fillna(0)  # internal stops remain, incur no stoplist length delta
        .groupby(["vehicle_id", "timestamp"])
        .sum()  # merge state changes at same time
        .groupby("vehicle_id")
        .cumsum()  # integrate over stoplist length changes for each vehicle
    )

    median_stoplist_length = stoplist_length_ensemble.median()
    mean_stoplist_length = stoplist_length_ensemble.mean()

    system_stoplist_length_ensemble = (
        event_log.reset_index("vehicle_id", drop=True)
        .loc[:, "event_type"]
        .map(dict(submission=2, pickup=-1, dropoff=-1))
        .fillna(0)
        .sort_index()
        .groupby("timestamp")
        .sum()
        .cumsum()
    )

    median_system_stoplist_length = system_stoplist_length_ensemble.median()
    mean_system_stoplist_length = system_stoplist_length_ensemble.mean()

    avg_detour = requests["inferred", "relative_travel_time"].mean()

    res = {}

    if params:
        n_vehicles = params["general"]["n_vehicles"] or len(
            params["general"]["initial_locations"]
        )
        space = params["general"]["space"]
        request_rate = params["request_generator"].get("rate")
        velocity = space.velocity
        d_avg = (
            sum(
                [
                    space.d(space.random_point(), space.random_point())
                    for _ in range(100_000)
                ]
            )
            / 100_000
        )

        load_requested = d_avg * request_rate / (velocity * n_vehicles)
        load_serviced = (
            d_avg * request_rate * (1 - rejection_ratio) / (velocity * n_vehicles)
        )

        res |= dict(
            n_vehicles=n_vehicles,
            request_rate=request_rate,
            velocity=velocity,
            load_requested=load_requested,
            load_serviced=load_serviced,
            avg_direct_dist_space=d_avg,
        )

    res |= dict(
        avg_occupancy=avg_occupancy,
        avg_segment_dist=avg_segment_dist,
        avg_segment_time=avg_segment_time,
        total_dist_driven=total_dist_driven,
        total_time_driven=total_time_driven,
        avg_direct_dist=avg_direct_dist,
        avg_direct_time=avg_direct_time,
        total_direct_dist=total_direct_dist,
        total_direct_time=total_direct_time,
        efficiency_dist=efficiency_dist,
        efficiency_time=efficiency_time,
        avg_waiting_time=avg_waiting_time,
        rejection_ratio=rejection_ratio,
        median_stoplist_length=median_stoplist_length,
        mean_stoplist_length=mean_stoplist_length,
        median_system_stoplist_length=median_system_stoplist_length,
        mean_system_stoplist_length=mean_system_stoplist_length,
        avg_detour=avg_detour,
    )

    _stops = stops.dropna(
        subset=[
            *(
                {
                    "system_stoplist_length_service_time",
                    "system_stoplist_length_submission_time",
                    "stoplist_length_service_time",
                    "stoplist_length_submission_time",
                }
                & set(stops)
            )
        ]
    )
    if "system_stoplist_length_service_time" in _stops:
        res["avg_system_stoplist_length_service_time"] = (
            _stops["system_stoplist_length_service_time"] * _stops["state_duration"]
        ).sum() / _stops["state_duration"].sum()

    if "system_stoplist_length_submission_time" in _stops:
        res["avg_system_stoplist_length_submission_time"] = (
            _stops["system_stoplist_length_submission_time"] * _stops["state_duration"]
        ).sum() / _stops["state_duration"].sum()

    if "stoplist_length_submission_time" in _stops:
        res["avg_stoplist_length_submission_time"] = (
            _stops["stoplist_length_submission_time"] * _stops["state_duration"]
        ).sum() / _stops["state_duration"].sum()

    if "stoplist_length_service_time" in _stops:
        res["avg_stoplist_length_service_time"] = (
            _stops["stoplist_length_service_time"] * _stops["state_duration"]
        ).sum() / _stops["state_duration"].sum()

    return res



import numpy as np
import pandas as pd


def get_vehicle_quantities(stops: pd.DataFrame, requests: pd.DataFrame) -> pd.DataFrame:
    """
    Compute various quantities aggregated **per vehicle**.

    Currently, the following observables are returned:

    - avg_occupancy
    - avg_segment_dist
    - avg_segment_time
    - total_dist_driven
    - total_time_driven
    - avg_direct_dist
    - avg_direct_time
    - total_direct_dist
    - total_direct_time
    - efficiency_dist
    - efficiency_time
    - avg_system_stoplist_length_service_time
    - avg_system_stoplist_length_submission_time
    - avg_stoplist_length_service_time
    - avg_stoplist_length_submission_time

    Parameters
    ----------
    stops
        Stops dataframe
    requests
        Requests dataframe

    Returns
    -------
    ``pd.DataFrame`` containing the aforementioned observables as columns, indexed by ``vehicle_id``
    """
    serviced_requests = (
        requests[requests[("rejected", "timestamp")].isna()]
        if ("rejected", "timestamp") in requests
        else requests
    )

    avg_occupancy = stops.groupby("vehicle_id").apply(
        lambda gdf: (gdf["occupancy"] * gdf["state_duration"]).sum()
        / gdf["state_duration"].sum()
    )

    avg_segment_dist = stops.groupby("vehicle_id")["dist_to_next"].mean()
    avg_segment_time = stops.groupby("vehicle_id")["time_to_next"].mean()

    total_dist_driven = stops.groupby("vehicle_id")["dist_to_next"].sum()
    total_time_driven = stops.groupby("vehicle_id")["time_to_next"].sum()

    if not serviced_requests.empty:
        avg_direct_dist = serviced_requests.groupby(
            ("serviced", "vehicle_id"), group_keys=False
        ).apply(lambda gdf: gdf.submitted.direct_travel_distance.mean())

        avg_direct_time = serviced_requests.groupby(
            ("serviced", "vehicle_id"), group_keys=False
        ).apply(lambda gdf: gdf.submitted.direct_travel_time.mean())

        total_direct_dist = serviced_requests.groupby(
            ("serviced", "vehicle_id"), group_keys=False
        ).apply(lambda gdf: gdf.submitted.direct_travel_distance.sum())

        total_direct_time = serviced_requests.groupby(
            ("serviced", "vehicle_id"), group_keys=False
        ).apply(lambda gdf: gdf.submitted.direct_travel_time.sum())

        efficiency_dist = total_direct_dist / total_dist_driven
        efficiency_time = total_direct_time / total_time_driven
    else:
        avg_direct_dist = pd.Series(np.nan, index=avg_occupancy.index, dtype=float)
        avg_direct_time = pd.Series(np.nan, index=avg_occupancy.index, dtype=float)
        total_direct_dist = pd.Series(np.nan, index=avg_occupancy.index, dtype=float)
        total_direct_time = pd.Series(np.nan, index=avg_occupancy.index, dtype=float)
        efficiency_dist = pd.Series(np.nan, index=avg_occupancy.index, dtype=float)
        efficiency_time = pd.Series(np.nan, index=avg_occupancy.index, dtype=float)

    res = pd.DataFrame(
        dict(
            avg_occupancy=avg_occupancy,
            avg_segment_dist=avg_segment_dist,
            avg_segment_time=avg_segment_time,
            total_dist_driven=total_dist_driven,
            total_time_driven=total_time_driven,
            avg_direct_dist=avg_direct_dist,
            avg_direct_time=avg_direct_time,
            total_direct_dist=total_direct_dist,
            total_direct_time=total_direct_time,
            efficiency_dist=efficiency_dist,
            efficiency_time=efficiency_time,
        )
    ).rename_axis("vehicle_id")

    if "system_stoplist_length_service_time" in stops:
        res["avg_system_stoplist_length_service_time"] = stops.groupby(
            "vehicle_id"
        ).apply(
            lambda gdf: (
                gdf["system_stoplist_length_service_time"] * gdf["state_duration"]
            ).sum()
            / gdf["state_duration"].sum()
        )
    if "system_stoplist_length_submission_time" in stops:
        res["avg_system_stoplist_length_submission_time"] = stops.groupby(
            "vehicle_id"
        ).apply(
            lambda gdf: (
                gdf["system_stoplist_length_submission_time"] * gdf["state_duration"]
            ).sum()
            / gdf["state_duration"].sum()
        )

    if "stoplist_length_service_time" in stops:
        res["avg_stoplist_length_service_time"] = stops.groupby("vehicle_id").apply(
            lambda gdf: (
                gdf["stoplist_length_service_time"] * gdf["state_duration"]
            ).sum()
            / gdf["state_duration"].sum()
        )

    if "stoplist_length_submission_time" in stops:
        res["avg_stoplist_length_submission_time"] = stops.groupby("vehicle_id").apply(
            lambda gdf: (
                gdf["stoplist_length_submission_time"] * gdf["state_duration"]
            ).sum()
            / gdf["state_duration"].sum()
        )

    return res



import os
import json
import shutil
import warnings
import subprocess

import functools as ft
import operator as op

from pathlib import Path

import loky
from tqdm import tqdm

from ridepy.extras.io import read_params_json, create_params_json, ParamsJSONDecoder
from ridepy.extras.simulation_set import SimulationSet, make_file_path
from ridepy.util import make_sim_id


def update_filenames(target_directory_path: Path):
    """
    With a directory containing simulation output (JSONL events files and params JSON files),
    this will try to update the simulation ids/base names to the id scheme of the current version of ridepy.

    The process is non-invasive in that it either creates symlinks or creates new files. The existing
    files are not renamed nor is their content modified.

    Parameters
    ----------
    target_directory_path
        Directory to operate in.
    """
    get_params_path = ft.partial(
        make_file_path, directory=target_directory_path, suffix="_params.json"
    )
    get_events_path = ft.partial(
        make_file_path, directory=target_directory_path, suffix=".jsonl"
    )

    old_ids = {
        str(p).split("/")[-1].split(".")[0].split("_")[0]
        for p in target_directory_path.glob("*.json")
        if not os.path.islink(p)
    }

    new_default_base_params = SimulationSet(data_dir=Path()).default_base_params

    for old_id in tqdm(old_ids):
        old_params_path = get_params_path(old_id)
        old_events_path = get_events_path(old_id)

        if not old_params_path.exists():
            warnings.warn(
                f"parameter file for simulation id {old_id} missing, skipping"
            )
            continue

        old_params_json = old_params_path.read_text(encoding="U8")

        params_json = (
            old_params_json.replace("thesimulator", "ridepy")
            .replace(
                "brute_force_total_traveltime_minimizing_dispatcher",
                "BruteForceTotalTravelTimeMinimizingDispatcher",
            )
            .replace(
                "simple_ellipse_dispatcher",
                "SimpleEllipseDispatcher",
            )
            .replace(
                "taxicab_dispatcher_drive_first",
                "TaxicabDispatcherDriveFirst",
            )
            .replace("TransportationRequestCls", "transportation_request_cls")
            .replace("VehicleStateCls", "vehicle_state_cls")
            .replace("FleetStateCls", "fleet_state_cls")
            .replace("RequestGeneratorCls", "request_generator_cls")
            .replace("dispatcher_class", "dispatcher_cls")
        )

        params = json.loads(params_json, cls=ParamsJSONDecoder)

        for outer_key, default_inner_dict in new_default_base_params.items():
            params[outer_key] = params.get(outer_key, default_inner_dict)
            for inner_key, inner_value in default_inner_dict.items():
                params[outer_key][inner_key] = params[outer_key].get(
                    inner_key, default_inner_dict.get(inner_key)
                )

        if "general" in params and "dispatcher" in params["general"]:
            params["dispatcher"]["dispatcher_cls"] = params["general"]["dispatcher"]
            del params["general"]["dispatcher"]

        new_params_json = create_params_json(params=params)
        new_id = make_sim_id(new_params_json)
        new_params_path = get_params_path(new_id)
        new_events_path = get_events_path(new_id)

        if new_params_json == old_params_json and old_params_path != new_params_path:
            try:
                os.symlink(old_params_path, new_params_path)
            except FileExistsError:
                warnings.warn(
                    f"parameter file for simulation id {old_id} already up to date, skipping"
                )
        else:
            new_params_path.write_text(new_params_json)

        if not old_events_path.exists():
            warnings.warn(f"{old_events_path} missing, skipping")
            continue
        elif old_events_path == new_events_path:
            warnings.warn(
                f"events file for simulation id {old_id} already up to date, skipping"
            )
        else:
            try:
                os.symlink(old_events_path, new_events_path)
            except FileExistsError:
                warnings.warn(
                    f"events file for simulation id {old_id} already up to date, skipping"
                )


def _update_events_file(
    sim_id, get_events_path, get_old_events_path, jq_inf, jq_options, jq_expr
):
    events_path = get_events_path(sim_id)
    old_events_path = get_old_events_path(sim_id)

    shutil.move(events_path, old_events_path)

    print(f"Reformatting {sim_id}...", end="", flush=True)
    try:
        subprocess.run(
            ["jq", *jq_options, jq_expr, str(old_events_path)],
            stdout=events_path.open("w"),
            check=True,
        )
        subprocess.run(
            ["sed", "-i", f"s/{jq_inf}/Infinity/g", f"{events_path}"], check=True
        )

        print("done.")
    except:
        print("FAILED.")
        shutil.move(old_events_path, events_path)
    else:
        old_events_path.unlink()


def update_events_files(target_directory_path: Path, max_workers=7):
    sim_ids = (
        events_path.stem
        for events_path in target_directory_path.glob("*.jsonl")
        if not os.path.islink(events_path)
    )

    get_events_path = ft.partial(
        make_file_path, directory=target_directory_path, suffix=".jsonl"
    )
    get_old_events_path = ft.partial(
        make_file_path, directory=target_directory_path, suffix="_old.jsonl"
    )

    jq_options = ["-r", "-c", "-e"]
    jq_expr = "{event_type:keys[0]}+.[keys[0]]"

    jq_inf = (
        subprocess.run(["jq", "."], input=b"Infinity", capture_output=True)
        .stdout.decode()
        .strip()
    )

    with loky.get_reusable_executor(max_workers=max_workers) as executor:
        list(
            executor.map(
                ft.partial(
                    _update_events_file,
                    get_events_path=get_events_path,
                    get_old_events_path=get_old_events_path,
                    jq_inf=jq_inf,
                    jq_options=jq_options,
                    jq_expr=jq_expr,
                ),
                sim_ids,
            )
        )



from __future__ import annotations

import datetime
import hashlib
import uuid
import random
import string
import sys
import warnings
from contextlib import contextmanager

import numpy as np
import dataclasses

from typing import Dict


MAX_SEAT_CAPACITY = sys.maxsize  # A very large int, because np.inf is a float


def get_short_uuid():
    return "".join(random.choices(string.ascii_lowercase + string.digits, k=8))


def get_uuid():
    return uuid.uuid4().hex


def get_datetime_yymmddHHSS():
    return datetime.datetime.now().strftime("%y%m%d%H%M")


class smartVectorize:
    """
    Method decorator for TransportSpace and its subclasses.

    Wraps methods to make them handle both operations
    on single space coordinates, and array-like bunches of coordinates.
    * checks whether dimensions of coordinates match with the space
    * loops over method handling single coordinates as arguments
    * dispatches to vectorized version of the function is existent

    Use like:
    ```
    @smartVectorize
    def foo(self, u, v):
        return magic(u, v)
    ```
    and optionally in addition
    ```
    @foo.vectorized
    def foo(self, u, v):
        return np.magic(u, v)
    ```
    """

    def __init__(self, base_fn, vectorized_fn=None, self_=None):
        # when the decorator is first applied to a class method,
        # only base_fn will be supplied and stored
        self.base_fn = base_fn
        self.vectorized_fn = vectorized_fn
        self.self_ = self_

    def __get__(self, obj, objtype=None):
        # when the decorated method is called, a reference to the
        # instance of the class that the decorated method belongs to
        # will be stored in smartVectorize().self_ to be able to later access
        # the classes n_dim attribute
        self.self_ = obj
        return self

    def __call__(self, *args, **kwargs):
        # When the decorated method is called, check the following things:
        # 1) If multiple args or kwargs or both are supplied, the shapes of the
        #    data supplied shall match. This may be up for debate, as hypothetically
        #    it could e.g. be useful to be able to supply both "vectors", i.e. coordinates,
        #    *and* scalars. Currently this is not implemented, though.

        shape = None

        # check homogenous shape for all positional arguments
        if args:
            shape = np.shape(args[0])
            if not all(np.shape(arg) == shape for arg in args[1:]):
                raise ValueError("vector shapes must match")

        # check homogenous shape for all keyword arguments and, if applicable,
        # make sure they also match the positional arguments' ones
        if kwargs:
            if shape is None:
                shape = np.shape(list(kwargs.values())[0])
            if not all(np.shape(v) == shape for v in kwargs.values()):
                raise ValueError("vector shapes must match")

        # 2) now determine whether we are dealing with a single coordinate per
        #    argument, or with multiple, i.e. an array of arguments

        # if the number of dimensions of the arguments matches
        # (mind that a single number here has dimensionality 0)
        # the number of dimensions of the transport space, we are dealing
        # with single coordinate vectors. Hence we call the base_fn normally.
        # NOTE: this may produce weird errors if the "correct" shape is not detected by np.shape, e.g.
        # for a pd.Series of coordinate tuples. In this case, the condition will evaluate true and feed
        # the series to base_fn
        if len(shape) == self.self_.n_dim - 1:
            return self.base_fn(self.self_, *args, **kwargs)
        # if however we have dimensionality of the space plus one,
        # we need to iterate over multiple coordinate vectors.
        elif len(shape) == self.self_.n_dim:
            # if there is no dedicated vectorized version, we need to use a for-loop
            if self.vectorized_fn is None:
                # in case we have only positional arguments, this is easy:
                if args and not kwargs:
                    res = [self.base_fn(self.self_, *arg) for arg in zip(*args)]
                elif not args and kwargs:
                    # if we only have keyword arguments, we need to unpack them accordingly
                    # and feed them to the base function which only takes single
                    # coordinate vectors:
                    res = [
                        self.base_fn(self.self_, **kwarg)
                        for kwarg in map(
                            dict,
                            zip(*[[(k, vv) for vv in v] for k, v in kwargs.items()]),
                        )
                    ]
                else:
                    # if we have both, we just zip both of the variants above together
                    # and continue as before
                    res = [
                        self.base_fn(self.self_, *arg, **kwarg)
                        for arg, kwarg in zip(
                            zip(*args),
                            map(
                                dict,
                                zip(
                                    *[[(k, vv) for vv in v] for k, v in kwargs.items()]
                                ),
                            ),
                        )
                    ]

                return np.array(res)

            # ...if however there is a dedicated vectorized function,
            #    we just forward everything and return
            else:
                return self.vectorized_fn(self.self_, *args, **kwargs)

        # If the dimensions of the input do not match the dimensions of the space, we complain.
        else:
            raise ValueError(
                f"Got {len(shape)}-dimensional object "
                f"instead of expected {self.self_.n_dim} dimensions"
            )

    def vectorized(self, vectorized_fn):
        # to add a vectorized implementation of the decorated method,
        # it is decorated with smartVectorize().vectorized, which adds a reference
        # to the vectorized method to the smartVectorize instance
        self.vectorized_fn = vectorized_fn
        return self


def make_dict(item, raise_errors: bool = True) -> Dict:
    """
    Convert data structure object to dict
    Parameters
    ----------
    item
        the object to convert
    raise_errors
        If true, raise TypeError

    Returns
    -------
    resulting dictionary

    """
    if dataclasses.is_dataclass(item):
        return dataclasses.asdict(item)
    elif hasattr(item, "asdict"):
        return item.asdict()
    elif raise_errors:
        raise TypeError(f"Cannot convert object of type {type(item)} to dict")


def make_repr(cls, dct):
    return f"{cls}(" + ", ".join((map(lambda s: f"{s[0]}={s[1]!r}", dct.items()))) + ")"


def make_sim_id(params_json: str):
    return hashlib.sha224(params_json.encode("ascii", errors="strict")).hexdigest()


@contextmanager
def supress_stoplist_extraction_warning():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            action="ignore",
            message=r"Extracting the C\+\+ stoplist will impact performance",
            category=UserWarning,
        )
        yield



from typing import Any, Union, TypedDict

from ridepy.data_structures import ID


class Event(TypedDict):
    """
    The base event class. Must hold a timestamp.

    .. code-block:: python

        event = {
            "event_type": "Event",
            "timestamp": ...,
        }

    """

    event_type: str
    timestamp: float


class RequestSubmissionEvent(Event):
    """
    Submission of a request with specific spatio-temporal constraints to the system.

    .. code-block:: python

        request_submission_event = {
            "event_type": "RequestSubmissionEvent",
            "timestamp": ...,
            "request_id": ...,
            "origin": ...,
            "destination": ...,
            "pickup_timewindow_min": ...,
            "pickup_timewindow_max": ...,
            "delivery_timewindow_min": ...,
            "delivery_timewindow_max": ...,
        }

    """

    request_id: ID
    origin: Any
    destination: Any
    pickup_timewindow_min: float
    pickup_timewindow_max: float
    delivery_timewindow_min: float
    delivery_timewindow_max: float


class RequestAcceptanceEvent(Event):
    """
    Commitment of the system to fulfil a request given
    the returned spatio-temporal constraints.

    .. code-block:: python

        request_acceptance_event = {
            "event_type": "RequestAcceptanceEvent",
            "timestamp": ...,
            "request_id": ...,
            "origin": ...,
            "destination": ...,
            "pickup_timewindow_min": ...,
            "pickup_timewindow_max": ...,
            "delivery_timewindow_min": ...,
            "delivery_timewindow_max": ...,
        }

    """

    request_id: ID
    origin: Any
    destination: Any
    pickup_timewindow_min: float
    pickup_timewindow_max: float
    delivery_timewindow_min: float
    delivery_timewindow_max: float


class RequestRejectionEvent(Event):
    """
    Inability of the system to fulfil a request.

    .. code-block:: python

        request_rejection_event = {
            "event_type": "RequestRejectionEvent",
            "timestamp": ...,
            "request_id": ...,
        }

    """

    request_id: ID


class PickupEvent(Event):
    """
    Successful pick-up action.

    .. code-block:: python

        pickup_event = {
            "event_type": "PickupEvent",
            "timestamp": ...,
            "request_id": ...,
            "vehicle_id": ...,
        }

    """

    request_id: ID
    vehicle_id: ID


class DeliveryEvent(Event):
    """
    Successful drop-off action.

    .. code-block:: python

        delivery_event = {
            "event_type": "DeliveryEvent",
            "timestamp": ...,
            "request_id": ...,
            "vehicle_id": ...,
        }

    """

    request_id: ID
    vehicle_id: ID


class InternalEvent(Event):
    """
    Successful internal action.

    .. code-block:: python

        internal_event = {
            "event_type": "InternalEvent",
            "timestamp": ...,
            "vehicle_id": ...,
        }

    """

    vehicle_id: ID


class VehicleStateBeginEvent(InternalEvent):
    """
    .. code-block:: python

        vehicle_state_begin_event = {
            "event_type": "VehicleStateBeginEvent",
            "timestamp": ...,
            "vehicle_id": ...,
            "location": ...,
            "request_id": -100,
        }

    """

    location: Any
    request_id: ID


class VehicleStateEndEvent(InternalEvent):
    """
    .. code-block:: python

        vehicle_state_end_event = {
            "event_type": "VehicleStateEndEvent",
            "timestamp": ...,
            "vehicle_id": ...,
            "location": ...,
            "request_id": -200,
        }

    """

    location: Any
    request_id: ID


RequestEvent = Union[
    RequestSubmissionEvent, RequestAcceptanceEvent, RequestRejectionEvent
]
"""Emitted when a `.TransportationRequest` is handled."""

StopEvent = Union[InternalEvent, PickupEvent, DeliveryEvent]
"""Emitted when a `.Stop` is serviced."""

RequestResponse = Union[RequestAcceptanceEvent, RequestRejectionEvent]
"""Emitted when a `.TransportationRequest` is handled."""



# distutils: language = c++

import random
import warnings
import itertools as it
import networkx as nx
import math as m

from libcpp.vector cimport vector
from libcpp.pair cimport pair
from libcpp.memory cimport dynamic_pointer_cast

from ridepy.util import smartVectorize
from .cspaces cimport(
    R2loc,
    Euclidean2D as CEuclidean2D,
    Manhattan2D as CManhattan2D,
    GraphSpace as CGraphSpace
)

from typing import List, Tuple, Optional

from ridepy.data_structures_cython.data_structures cimport LocType, R2loc
from ridepy.util.spaces import Euclidean2D as pyEuclidean2D

from cython.operator cimport dereference

"""
Note: We are duplicating the C++ class hierarchy in ./spaces_cython.h. In short, our C++ transport spaces
all inherit from the abstract base class TransportSpace. Here, we will create a cdef class also called
TransportSpace and other cdef classes will inherit from that. There are a few caveats, which were described
in great length at https://stackoverflow.com/a/28727488. Basically:

1. Virtual functions that are later overridden in a child class **must be declared only for the base class**.
   That is, we must *not* declare Euclidean2D.d/t/... here, only in TransportSpace.
   Doing so will result in an *ambiguous overridden function* error from cython.
2. When allocating/deallocating c pointers in __cinit__/__dealloc__ of a cdef class ClassLaterSubclassedFrom 
   that has been subclassed, a check like `if type(self) == ClassLaterSubclassedFrom` must be used. Otherwise
   double free occurs. 
"""
cdef class TransportSpace:
    """
    Base class for extension types wrapping C++ TransportSpace class template. Since there's no elegant way of
    wrapping templates in cython and exposing them to python, we will use the [Explicit Run-Time Dispatch approach]
    (https://martinralbrecht.wordpress.com/2017/07/23/adventures-in-cython-templating/). See the docstring of
    ridepy/data_structures_cython/data_structures.pyx for details.
    """
    def __init__(self, loc_type):
        if loc_type == LocType.INT:
            self.loc_type = LocType.INT
            self.n_dim = 1
        elif loc_type == LocType.R2LOC:
            self.loc_type = LocType.R2LOC
            self.n_dim = 2
        else:
            raise ValueError("This line should never have been reached")

    @smartVectorize
    def d(self, u, v):
        if self.loc_type == LocType.R2LOC:
            return dereference(self.u_space.space_r2loc_ptr).d(<R2loc>u, <R2loc>v)
        elif self.loc_type == LocType.INT:
            return dereference(self.u_space.space_int_ptr).d(<int>u, <int>v)
        else:
            raise ValueError("This line should never have been reached")

    @smartVectorize # TODO: check if the smartVectorize works for this cdef class
    def t(self, u, v):
        if self.loc_type == LocType.R2LOC:
            return dereference(self.u_space.space_r2loc_ptr).t(<R2loc>u, <R2loc>v)
        elif self.loc_type == LocType.INT:
            return dereference(self.u_space.space_int_ptr).t(<int>u, <int>v)

    def interp_dist(self, u, v, double dist_to_dest):
        if self.loc_type == LocType.R2LOC:
            return dereference(self.u_space.space_r2loc_ptr).interp_dist(<R2loc>u, <R2loc>v, dist_to_dest)
        elif self.loc_type == LocType.INT:
            return dereference(self.u_space.space_int_ptr).interp_dist(<int>u, <int>v, dist_to_dest)
        else:
            raise ValueError("This line should never have been reached")

    def interp_time(self, u, v, double time_to_dest):
        if self.loc_type == LocType.R2LOC:
            return dereference(self.u_space.space_r2loc_ptr).interp_time(<R2loc>u, <R2loc>v, time_to_dest)
        elif self.loc_type == LocType.INT:
            return dereference(self.u_space.space_int_ptr).interp_time(<int>u, <int>v, time_to_dest)
        else:
            raise ValueError("This line should never have been reached")

    def asdict(self):
        return dict(loc_type=self.loc_type)

    def __dealloc__(self):
        """
        Since this is a base class that will never be instantiated, we do not need to free any pointer here.
        Take care to do so in the derived classes though.
        """
        ...

    def __eq__(self, other: TransportSpace):
        return type(self) == type(other) and self.asdict() == other.asdict()


cdef class Euclidean2D(TransportSpace):
    """
    :math:`\mathbb{R}^2` with :math:`L_2`-induced metric (Euclidean).
    """
    def __cinit__(self, double velocity=1, *args, **kwargs):
        self.loc_type = LocType.R2LOC
        self.derived_ptr = self.u_space.space_r2loc_ptr = new CEuclidean2D(velocity)

    def __init__(self, *args, **kwargs): # remember both __cinit__ and __init__ gets the same arguments passed
        """
        Parameters
        ----------
        velocity
            constant velocity to compute travel time, optional. default: 1
        """
        TransportSpace.__init__(self, loc_type=LocType.R2LOC)
        coord_range = kwargs.get('coord_range') or (args[1] if len(args) > 1 else None)

        if coord_range is not None:
            assert len(coord_range) == self.n_dim, (
                "Number of desired dimensions must "
                "match the number of coord range pairs given"
            )
            self.coord_range = coord_range
        else:
            self.coord_range = [(0, 1)] * self.n_dim

    def __dealloc__(self):
        del self.derived_ptr

    def __repr__(self):
        return f"Euclidean2D(velocity={self.velocity})"

    @property
    def velocity(self):
        return dereference(self.derived_ptr).velocity

    def random_point(self):
        return tuple(random.uniform(a, b) for a, b in self.coord_range)

    def asdict(self):
        return dict(velocity=self.velocity, coord_range=self.coord_range)

    def __reduce__(self):
        return self.__class__, (self.velocity, self.coord_range)


cdef class Manhattan2D(TransportSpace):
    """
    :math:`\mathbb{R}^2` with :math:`L_1`-induced metric (Manhattan).
    """
    def __cinit__(self, double velocity=1, *args, **kwargs):
        self.loc_type = LocType.R2LOC
        self.derived_ptr = self.u_space.space_r2loc_ptr = new CManhattan2D(velocity)

    def __init__(self, *args, **kwargs): # remember both __cinit__ and __init__ gets the same arguments passed
        """
        Parameters
        ----------
        velocity
            constant velocity to compute travel time, optional. default: 1
        """
        TransportSpace.__init__(self, loc_type=LocType.R2LOC)

        coord_range = kwargs.get('coord_range') or (args[1] if len(args) > 1 else None)
        if coord_range is not None:
            assert len(coord_range) == self.n_dim, (
                "Number of desired dimensions must "
                "match the number of coord range pairs given"
            )
            self.coord_range = coord_range
        else:
            self.coord_range = [(0, 1)] * self.n_dim

    def __dealloc__(self):
        del self.derived_ptr

    def __repr__(self):
        return f"Manhattan2D(velocity={self.velocity})"

    @property
    def velocity(self):
        return dereference(self.derived_ptr).velocity

    def random_point(self):
        return tuple(random.uniform(a, b) for a, b in self.coord_range)

    def asdict(self):
        return dict(velocity=self.velocity, coord_range=self.coord_range)

    def __reduce__(self):
        return self.__class__, (self.velocity, self.coord_range)


cdef class Grid2D(Manhattan2D):
    """
    Fast 2D grid network space, realized as a discrete 2D space with Manhattan metric.

    Note that this space does not enforce the discrete integer coordinate pairs it is
    supposed to be used with. This means that you have to take care yourself to only
    introduce integer coordinate pairs. This implies that the distance and time functions
    ``Grid2D.d`` and ``Grid2D.t`` will in fact as if the space were continuous.
    The discrete/graph characteristic of the space is encoded in the interpolation and
    random point generation functions.

    The interpolation function will always assume the shortest path with no turns or a
    single right turn.
    """
    def __cinit__(self, int n=10, int m=10, double dn=1, double dm=1, double velocity=1):
        self.loc_type = LocType.R2LOC
        self.derived_ptr = self.u_space.space_r2loc_ptr = new CManhattan2D(velocity)

    def __init__(self, n=10, m=10, dn=1, dm=1, velocity=1):
        """
        Parameters
        ----------
        velocity
            constant velocity to compute travel time, optional. default: 1
        """
        TransportSpace.__init__(self, loc_type=LocType.R2LOC)

        self.n = n
        self.m = m
        self.dn = dn
        self.dm = dm
        self.velocity = velocity

    def d(self, u, v):
        return abs(u[0] - v[0]) * self.dn + abs(u[1] - v[1]) * self.dm

    def t(self, u, v):
        return self.d(u, v) / self.velocity

    def interp_dist(self, u, v, double dist_to_dest):
        d_r = dist_to_dest # Remaining distance
        d_t = self.d(u, v) # Total distance
        d_e = d_t - d_r # Elapsed distance
        d_n = self.dn # Vertical grid spacing
        d_m = self.dm # Horizontal grid spacing


        if d_r == d_t:
            # We are still at the origin
            # print("We are still at the origin")
            i = u[0]
            j = u[1]
            d_j = 0
        elif d_r == 0:
            # We have reached the destination
            # print("We have reached the destination")
            i = v[0]
            j = v[1]
            d_j = 0
        else:
            # We are under way
            # print("We are under way")
            w = (v[0] - u[0], v[1] - u[1])

            d_vert = abs(w[0]) * d_n
            d_hori = abs(w[1]) * d_m

            if (w[0] < 0 < w[1]) or (w[0] > 0 > w[1]) or (w[1] == 0):
                # Going vertically, first
                # print("Going vertically, first")
                if d_e <= d_vert:
                    # We have not made the turn.
                    # print("We have not made the turn")
                    j = u[1]
                    # Determine next node:
                    if w[0] < 0:
                        # Going upwards
                        # print("Going upwards")
                        d_e_d = d_e
                        k_d_e_d = m.ceil(d_e_d / d_n)
                        i = u[0] - k_d_e_d
                        d_j = k_d_e_d * d_n - d_e_d
                    else:
                        # Going downwards
                        # print("Going downwards")
                        d_e_d = d_e
                        k_d_e_d = m.ceil(d_e_d / d_n)
                        i = u[0] + k_d_e_d
                        d_j = k_d_e_d * d_n - d_e_d
                else:
                    # We have made the turn.
                    # print("We have made the turn")
                    i = v[0]
                    # Determine next node:
                    if w[1] < 0:
                        # Going left
                        # print("Going left")
                        d_e_d = d_e - d_vert
                        k_d_e_d = m.ceil(d_e_d / d_m)
                        j = u[1] - k_d_e_d
                        d_j =  k_d_e_d * d_m - d_e_d
                    else:
                        # Going right
                        # print("Going right")
                        d_e_d = d_e - d_vert
                        k_d_e_d = m.ceil(d_e_d / d_m)
                        j = u[1] + k_d_e_d
                        d_j =  k_d_e_d * d_m - d_e_d
            else:
                # Going horizontally, first
                # print("Going horizontally, first")
                if d_e <= d_hori:
                    # We have not made the turn
                    # print("We have not made the turn")
                    i = u[0]
                    # Determine the next node:
                    if w[1] < 0:
                        # Going left
                        # print("Going left")
                        d_e_d = d_e
                        k_d_e_d = m.ceil(d_e_d / d_m)
                        j = u[1] - k_d_e_d
                        d_j =  k_d_e_d * d_m - d_e_d
                    else:
                        # Going right
                        # print("Going right")
                        d_e_d = d_e
                        k_d_e_d = m.ceil(d_e_d / d_m)
                        j = u[1] + k_d_e_d
                        d_j =  k_d_e_d * d_m - d_e_d
                else:
                    # We have made the turn
                    # print("We have made the turn")
                    j = v[1]
                    # Determine the next node:
                    if w[0] < 0:
                        # Going upwards
                        # print("Going upwards")
                        d_e_d = d_e - d_hori
                        k_d_e_d = m.ceil(d_e_d / d_n)
                        i = u[0] - k_d_e_d
                        d_j = k_d_e_d * d_n - d_e_d
                    else:
                        # Going downwards
                        # print("Going downwards")
                        d_e_d = d_e - d_hori
                        k_d_e_d = m.ceil(d_e_d / d_n)
                        i = u[0] + k_d_e_d
                        d_j = k_d_e_d * d_n - d_e_d

            print(f"d_r = {d_r}, d_t = {d_t}, d_e = {d_e}, d_n = {d_n}, d_m = {d_m}, w = {w}, d_vert = {d_vert}, "
                  f"d_hori = {d_hori}, d_e_d = {d_e_d}, k_d_e_d = {k_d_e_d}, i = {i}, j = {j}, d_j = {d_j}")
        return (round(i), round(j)), d_j


    def interp_time(self, u, v, double time_to_dest):
        nn, d_j = self.interp_dist(u, v, time_to_dest * self.velocity)
        t_j = d_j / self.velocity
        return nn, t_j

    def random_point(self):
        i = random.randint(0, self.n - 1)
        j = random.randint(0, self.m - 1)
        return i, j

    def __repr__(self):
        return f"Grid2D(n={self.n}, m={self.m}, dn={self.dn}, dm={self.dm}, velocity={self.velocity})"

    def asdict(self):
        return dict(n=self.n, m=self.m, dn=self.dn, dm=self.dm, velocity=self.velocity)
    def __reduce__(self):
        return self.__class__, (self.n, self.m, self.dn, self.dm, self.velocity)

cdef class Grid2D_QM(Grid2D):
    """
    Same as Grid2D, except that the bus routes are nonunique. This means that the choice of the
    (non-unique) shortest path on the grid will only be made at the time of interpolation ("measurement").
    This has the benefit that buses tend to move along the Euclidean shortest path. The drawback is that
    interpolating multiple times along the route will produce inconsistent paths taken.
    """

    def interp_dist(self, u, v, double dist_to_dest):
        d_r = dist_to_dest
        d_n = self.dn
        d_m = self.dm

        if d_r == 0:
            nn = v
            d_j = 0
        else:
            frac = d_r / self.d(u, v)
            x = (u[0] * frac + (1 - frac) * v[0]) * d_n, (u[1] * frac + (1 - frac) * v[1]) * d_m
            # print(x)

            # determine grid cell target node:
            if u[0] < v[0]:
                # going downwards
                # print('going down')
                i = m.ceil(x[0] / d_n)
            elif u[0] > v[0]:
                # going upwards
                # print('going up')
                i = m.floor(x[0] / d_n)
            else:
                # going horizontally
                i = u[0] / d_n

            if u[1] < v[1]:
                # going right
                # print('going right')
                j = m.ceil(x[1] / d_m)
            elif u[1] > v[1]:
                # going left
                # print('going left')
                j = m.floor(x[1] / d_m)
            else:
                # going vertically
                j = u[1] / d_m

            # distance remaining from grid cell target node
            d_r_p = abs(v[0] - i) * d_n + abs(v[1] - j) * d_m

            # distance within grid cell to grid cell target node
            d_g = d_r - d_r_p

            # assumption: we always deviate to the left
            if (u[0] >= v[0] and u[1] < v[1]) or (u[0] <= v[0] and u[1] > v[1]):
                # coming in horizontally
                # print('coming in horizontally')
                horizontal = True
                d_gref = d_m
            else:
                # coming in vertically
                # print('coming in vertically')
                horizontal = False
                d_gref = d_n

            if d_g < d_gref:
                # next node is grid cell target node,
                # jump distance is grid cell distance
                nn = (i, j)
                d_j = d_g
            else:
                d_j = d_g - d_gref
                if horizontal:
                    # coming in horizontally
                    if u[1] < v[1]:
                        # coming in from the left
                        # print('coming in from the left')
                        nn = (i, j - 1)
                    else:
                        # coming in from the right
                        # print('coming in from the right')
                        nn = (i, j + 1)
                else:
                    # coming in vertically
                    if u[0] < v[0]:
                        # coming in from above
                        # print('coming in from above')
                        nn = (i - 1, j)
                    else:
                        # coming in from below
                        # print('coming in from below')
                        nn = (i + 1, j)

        return (round(nn[0]), round(nn[1])), d_j


    def __repr__(self):
        return f"Grid2D_QM(n={self.n}, m={self.m}, dn={self.dn}, dm={self.dm}, velocity={self.velocity})"

cdef class Graph(TransportSpace):
    """
    Weighted directed graph with integer node labels.
    """
    def __cinit__(self, vertices, edges, weights=None, double velocity=1):
        self.loc_type = LocType.INT

        if weights is None:
            self.derived_ptr = self.u_space.space_int_ptr = new CGraphSpace[int](
                velocity, <vector[int]>vertices, <vector[pair[int, int]]>edges
            )
        else:
            if isinstance(weights, (int, float)):
                weights = it.repeat(float(weights), len(edges))

            self.derived_ptr = self.u_space.space_int_ptr = new CGraphSpace[int](
                velocity, <vector[int]>vertices, <vector[pair[int, int]]>edges, <vector[double]>weights
            )

    def __init__(self, *args, **kwargs): # remember both __cinit__ and __init__ gets the same arguments passed
        """
        Parameters
        ----------
        vertices : Sequence[int]
            sequence of vertices
        edges : Sequence[Tuple[int, int]]
            sequence of edge tuples
        weights : Union[None, float, Sequence[float]]
            Edge weights.
            - if None is supplied, the resulting graph is unweighted (unit edge length)
            - if a single float is supplied, every edge length will be equal to this number
            - if a sequence is supplied, this will be mapped onto the edge sequence
        velocity
            constant velocity to compute travel time, optional. default: 1
        """
        TransportSpace.__init__(self, loc_type=LocType.INT)

    def __dealloc__(self):
        del self.derived_ptr

    def __repr__(self):
        return f"Graph(velocity={self.velocity})"

    @property
    def velocity(self):
        return dereference(self.derived_ptr).velocity

    @property
    def vertices(self):
        return dereference(self.derived_ptr).get_vertices()

    @property
    def edges(self):
        return dereference(self.derived_ptr).get_edges()

    @property
    def weights(self):
        return dereference(self.derived_ptr).get_weights()

    @classmethod
    def from_nx(
            cls, G, velocity: float = 1.0, make_attribute_distance: Optional[str] = "distance"
    ):
        """
        Create a Graph from a networkx.Graph with a mandatory distance edge attribute.

        Parameters
        ----------
        G : networkx.Graph
            networkx graph
        velocity
            velocity used for travel time computation
        make_attribute_distance
            name of the nx.DiGraph edge attribute used as distance between nodes.
            If None is supplied, the resulting graph is unweighted (unit edge length).

        Returns
        -------
        Graph instance
        """
        if not isinstance(G, nx.Graph):
            raise TypeError(f"Must supply nx.Graph, not {type(G)}")
        elif G.is_directed():
            raise TypeError(f"Must supply undirected graph")

        if not all(isinstance(u, int) for u in random.sample(list(G.nodes()), k=min(5, len(G)))):
            warnings.warn("Heuristic determined non-int node labels. Converting to int", UserWarning)
            G = nx.relabel.convert_node_labels_to_integers(G)

        if make_attribute_distance is None:
            weights = None
        else:
            weights = [G[u][v][make_attribute_distance] for u, v in G.edges()]

        return cls(
            vertices=list(G.nodes()),
            edges=list(G.edges()),
            weights=weights,
            velocity=velocity
        )

    def random_point(self):
        return random.choice(self.vertices)


    def __reduce__(self):
        return self.__class__, \
            (
                 self.vertices,
                 self.edges,
                 self.weights,
                 self.velocity,
            )

    def asdict(self):
        return dict(
            vertices=self.vertices,
            edges=self.edges,
            weights=self.weights,
            velocity=self.velocity,
        )



import numpy as np

from ridepy.data_structures import (
    TransportationRequest,
    Stoplist,
    TransportSpace,
    DispatcherSolution,
    Stop,
    StopAction,
)
from ridepy.util.dispatchers.dispatcher_class import dispatcherclass


@dispatcherclass
def TaxicabDispatcherDriveFirst(
    request: TransportationRequest,
    stoplist: Stoplist,
    space: TransportSpace,
    seat_capacity: int,
) -> DispatcherSolution:
    """
    Dispatcher that maps a vehicle's stoplist and a request to a new stoplist
    by simply appending the necessary stops to the existing stoplist.

    This pure function is turned into a callable class by the decorator `.dispatcherclass` whose __init__ accepts
    optional arguments, see the docstring of `.dispatcherclass` for details.

    See the dispatcher interface in :ref:`desc_dispatcher` for details.

    Parameters
    ----------
    request
        request to be serviced.
    stoplist
        stoplist of the vehicle, to be mapped to a new stoplist.
    space
        transport space the vehicle is operating on.
    seat_capacity
        the maximum number of `.TransportationRequest` s that can be in a vehicle at the same time.


    Returns
    -------
        The best solution as defined in `.SingleVehicleSolution`.
    """
    # TODO: When we have multi-passenger requests, this dispatcher needs to be changed and
    # include capacity constraints. Currently, taxi := single seat
    assert seat_capacity == 1
    CPAT_pu = max(
        stoplist[-1].estimated_arrival_time,
        stoplist[-1].time_window_min if stoplist[-1].time_window_min is not None else 0,
    ) + space.t(stoplist[-1].location, request.origin)
    EAST_pu = request.pickup_timewindow_min
    CPAT_do = max(EAST_pu, CPAT_pu) + space.t(request.origin, request.destination)
    LAST_pu = (
        CPAT_pu + request.delivery_timewindow_max
        if request.delivery_timewindow_max is not None
        else np.inf
    )
    EAST_do = EAST_pu
    LAST_do = np.inf

    cost = CPAT_do
    stoplist = stoplist + [
        Stop(
            location=request.origin,
            request=request,
            action=StopAction.pickup,
            estimated_arrival_time=CPAT_pu,
            occupancy_after_servicing=stoplist[-1].occupancy_after_servicing + 1,
            time_window_min=EAST_pu,
            time_window_max=LAST_pu,
        ),
        Stop(
            location=request.destination,
            request=request,
            action=StopAction.dropoff,
            estimated_arrival_time=CPAT_do,
            occupancy_after_servicing=0,
            time_window_min=EAST_do,
            time_window_max=LAST_do,
        ),
    ]
    return cost, stoplist, (EAST_pu, LAST_pu, EAST_do, LAST_do)



import functools as ft
from collections import defaultdict

import numpy as np
import pandas as pd


def _create_stoplist_dataframe(*, evs: pd.DataFrame) -> pd.DataFrame:
    """
    Creates a DataFrame containing the stoplists of the transporters.
    The location is only returned for the internal stops relating to
    VehicleStateBeginEvent and VehicleStateEndEvent. For the rest of
    the stops, location is nan.

    The schema of the returned DataFrame is the following, where
    `vehicle_id` and `timestamp` are MultiIndex levels.

    .. code-block::

        Column           Dtype
        ------           -----
        vehicle_id       float64
        timestamp        float64
        delta_occupancy  float64
        request_id       object
        state_duration   float64
        occupancy        float64
        location         Union[float64, int, Tuple[float64]]

    Parameters
    ----------
    evs
        events DataFrame

    Returns
    -------
    stoplist DataFrame indexed by `vehicle_id` and `timestamp`
    """
    # Create the initial dataframe: Select only events of type
    # `PickupEvent` and `DeliveryEvent`, along with their vehicle_id, timestamp,
    # and request_id
    stops = evs[
        evs["event_type"].isin(
            [
                "PickupEvent",
                "DeliveryEvent",
                "VehicleStateBeginEvent",
                "VehicleStateEndEvent",
            ]
        )
    ][["vehicle_id", "timestamp", "event_type", "request_id", "location"]]

    # Translate PickupEvent and DeliveryEvent into an occupancy delta.
    # NOTE: This assumes occupancy delta of +1/-1, i.e. only single-customer requests.
    #       If the simulator should allow for multi-customer requests in the future,
    #       this must be changed.
    #       See also [issue #45](https://github.com/PhysicsOfMobility/ridepy/issues/45)
    stops["delta_occupancy"] = stops["event_type"].map(
        defaultdict(float, PickupEvent=1, DeliveryEvent=-1)
    )

    # ... and drop the event_type as pickup/dropoff is now signified through delta_occupancy
    stops.drop("event_type", axis=1, inplace=True)

    stops.sort_values(["vehicle_id", "timestamp", "request_id"], inplace=True)

    # Fix the stop order. The begin and end stops have identical timestamps as
    # other stops, partially on the same vehicle. This is problematic as for
    # proper computation of the state durations BEGIN and END **must** be first
    # and last stops in every stoplist.
    def fix_start_stop_order(df):
        # get absolute current positions of the BEGIN/END stops
        i_start = (df["request_id"] == -100).argmax()
        i_stop = (df["request_id"] == -200).argmax()

        # get dataframe's integer index values for the dummy stops
        idx = df.index.to_list()

        k_start = idx[i_start]
        k_stop = idx[i_stop]

        # delete the existing stops from the index list...
        if i_start < i_stop:
            i_stop -= 1
        else:
            i_start -= 1

        del idx[i_start]
        del idx[i_stop]

        # ...and insert them at the correct positions
        idx.insert(0, k_start)
        idx.append(k_stop)

        return df.loc[idx]

    stops = stops.groupby("vehicle_id", as_index=False, group_keys=False).apply(
        fix_start_stop_order
    )

    # compute the durations of every state and add them as a columns to the dataframe
    stops["state_duration"] = (
        stops.groupby("vehicle_id")["timestamp"].diff().shift(-1).fillna(0)
    )
    # compute the occupancy as delta_occupancy cumsum
    stops["occupancy"] = stops.groupby("vehicle_id")["delta_occupancy"].cumsum()

    # set index to ('vehicle_id, 'stop_id'), where stop_id in 0...N for each vehicle
    stops = (
        stops.groupby("vehicle_id")
        .apply(lambda df: df.reset_index(drop=True))
        .drop("vehicle_id", axis=1)
    )
    stops.index.rename(names="stop_id", level=1, inplace=True)

    # check total operational times of all vehicles are almost identical
    iterator = iter(stops.groupby("vehicle_id")["state_duration"].sum())
    try:
        first = next(iterator)
    except StopIteration:
        pass
    else:
        assert all(np.isclose(first, x) for x in iterator)

    return stops


def _add_locations_to_stoplist_dataframe(*, reqs, stops, space) -> pd.DataFrame:
    """
    Add non-internal stops' locations to the stoplist DataFrame as inferred
    from the *accepted* requests.

    The schema of the returned DataFrame is a superset of the following, where
    `vehicle_id` and `timestamp` are MultiIndex levels.

    .. code-block::

        Column           Dtype
        ------           -----
        vehicle_id       float64
        timestamp        float64
        delta_occupancy  float64
        request_id       object
        state_duration   float64
        occupancy        float64
        location         Union[float64, int, Tuple[float64]]
        time_to_next     float64
        dist_to_next     float64

    Parameters
    ----------
    reqs
        requests  DataFrame
    stops
        stops DataFrame missing stop locations for non-internal stops

    Returns
    -------
    stoplist DataFrame with added stop locations indexed by `vehicle_id` and `timestamp`
    """

    # use the requests' locations and reshape them into a DateFrame indexed by
    # `request_id` and `delta_occupancy`
    if "accepted" in reqs.columns:
        locations = reqs.loc[:, ("accepted", ["origin", "destination"])]
    else:
        locations = pd.DataFrame(
            index=reqs.index,
            columns=pd.MultiIndex.from_product(
                [["accepted"], ["origin", "destination"]]
            ),
        )
    locations.columns = locations.columns.droplevel(0).rename("delta_occupancy")
    locations = locations.stack().rename("location")

    # NOTE: This assumes occupancy delta of +1/-1, i.e. only single-customer requests.
    #       If the simulator should allow for multi-customer requests in the future,
    #       this must be changed.
    #       See also [issue #45](https://github.com/PhysicsOfMobility/ridepy/issues/45)
    locations.index = locations.index.set_levels(
        levels=locations.index.levels[1].map({"origin": 1.0, "destination": -1.0}),
        level=1,
    )

    # finally fill the locations missing in the stops dataframe by joining on request_id and delta_occupancy
    # and subsequently filling up nan from the now index-aligned "location_tmp" series
    stops["location"] = stops["location"].fillna(
        stops.join(locations, on=["request_id", "delta_occupancy"], rsuffix="_tmp")[
            "location_tmp"
        ],
    )

    def dist_time_to_next(df):
        locs = df["location"]

        dist_to_next = space.d(locs[:-1].to_list(), locs[1:].to_list())
        df["dist_to_next"] = pd.Series(dist_to_next, index=df.index[:-1])

        time_to_next = space.t(locs[:-1].to_list(), locs[1:].to_list())
        df["time_to_next"] = pd.Series(time_to_next, index=df.index[:-1])

        return df

    stops = stops.groupby("vehicle_id", group_keys=False).apply(dist_time_to_next)

    return stops[
        [
            "timestamp",
            "delta_occupancy",
            "request_id",
            "state_duration",
            "occupancy",
            "location",
            "dist_to_next",
            "time_to_next",
        ]
    ]


def _add_insertion_stats_to_stoplist_dataframe(*, reqs, stops, space) -> pd.DataFrame:
    """

    The schema of the returned DataFrame is a superset of the following, where
    `vehicle_id` and `timestamp` are MultiIndex levels.

    .. code-block::

        Column                                       Dtype
        ------                                       -----
        vehicle_id                                 float64
        stop_id                                      int64
        timestamp                                  float64
        delta_occupancy                            float64
        request_id                                   int64
        state_duration                             float64
        occupancy                                  float64
        location                                    object
        dist_to_next                               float64
        time_to_next                               float64
        timestamp_submitted                        float64
        insertion_index                            float64
        leg_1_dist_service_time                    float64
        leg_2_dist_service_time                    float64
        leg_direct_dist_service_time               float64
        detour_dist_service_time                   float64
        leg_1_dist_submission_time                 float64
        leg_2_dist_submission_time                 float64
        leg_direct_dist_submission_time            float64
        detour_dist_submission_time                float64
        stoplist_length_submission_time            float64
        stoplist_length_service_time               float64
        avg_segment_dist_submission_time           float64
        avg_segment_time_submission_time           float64
        avg_segment_dist_service_time              float64
        avg_segment_time_service_time              float64
        system_stoplist_length_submission_time     float64
        system_stoplist_length_service_time        float64
        avg_system_segment_dist_submission_time    float64
        avg_system_segment_time_submission_time    float64
        avg_system_segment_dist_service_time       float64
        avg_system_segment_time_service_time       float64
        relative_insertion_position                float64



    Parameters
    ----------
    reqs
        requests  DataFrame
    stops
        stops DataFrame

    Returns
    -------
    stoplist DataFrame
    """

    stops = pd.merge(
        stops,
        reqs.submitted.timestamp,
        how="left",
        left_on="request_id",
        right_index=True,
        suffixes=(None, "_submitted"),
    )
    actual_stops = stops.dropna(subset=("timestamp_submitted",))

    def _properties_at_time(stop, full_sl, scope):
        t = stop["timestamp"]
        ts = stop["timestamp_submitted"]
        pu = True if stop["delta_occupancy"] > 0 else False

        get_i_pu = lambda _sl, stop: (_sl["request_id"] == stop["request_id"]).argmax()
        get_i_do = (
            lambda _sl, stop: len(_sl)
            - (_sl["request_id"] == stop["request_id"])[::-1].argmax()
            - 1
        )

        sl = full_sl[
            (full_sl["timestamp_submitted"] <= t) & (t <= full_sl["timestamp"])
        ]

        sl_s = full_sl[
            (full_sl["timestamp_submitted"] <= ts) & (ts <= full_sl["timestamp"])
        ]

        if pu:
            i_pu_sl = get_i_pu(sl, stop)
            i_pu_sl_s = get_i_pu(sl_s, stop)
            i_do_sl_s = get_i_do(sl_s, stop)

            idx_pu = sl_s.iloc[i_pu_sl_s].name
            idx_do = sl_s.iloc[i_do_sl_s].name
            assert idx_pu == stop.name

            i_stop_sl = i_pu_sl
            i_stop_sl_s = i_pu_sl_s
        else:
            i_do_sl = get_i_do(sl, stop)

            i_pu_sl_s = get_i_pu(sl_s, stop)
            i_do_sl_s = get_i_do(sl_s, stop)

            idx_pu = sl_s.iloc[i_pu_sl_s].name
            idx_do = sl_s.iloc[i_do_sl_s].name
            assert idx_do == stop.name

            i_stop_sl = i_do_sl
            i_stop_sl_s = i_do_sl_s

        res = {}

        if scope != "system":
            res["insertion_index"] = len(sl_s[sl_s["timestamp"] < t])

            def _get_legs(i_stop, _sl, time_desc):
                l_sl = len(_sl)
                if i_stop == 0 == l_sl - 1:
                    stop_leg_1 = 0
                    stop_leg_2 = 0
                    stop_leg_d = 0
                elif i_stop == 0:
                    stop_leg_1 = 0
                    stop_leg_2 = space.d(
                        stop["location"], _sl.iloc[i_stop + 1]["location"]
                    )
                    stop_leg_d = stop_leg_1
                elif i_stop == l_sl - 1:
                    stop_leg_1 = space.d(
                        _sl.iloc[i_stop - 1]["location"], stop["location"]
                    )
                    stop_leg_2 = 0
                    stop_leg_d = stop_leg_2
                else:
                    stop_leg_1 = space.d(
                        _sl.iloc[i_stop - 1]["location"], stop["location"]
                    )
                    stop_leg_2 = space.d(
                        stop["location"], _sl.iloc[i_stop + 1]["location"]
                    )
                    stop_leg_d = space.d(
                        _sl.iloc[i_stop - 1]["location"],
                        _sl.iloc[i_stop + 1]["location"],
                    )

                stop_detour = stop_leg_1 + stop_leg_2 - stop_leg_d

                return {
                    f"leg_1_dist_{time_desc}_time": stop_leg_1,
                    f"leg_2_dist_{time_desc}_time": stop_leg_2,
                    f"leg_direct_dist_{time_desc}_time": stop_leg_d,
                    f"detour_dist_{time_desc}_time": stop_detour,
                }

            res |= _get_legs(i_stop_sl, sl, "service")
            res |= _get_legs(i_stop_sl_s, sl_s, "submission")

        if pu:
            sl.drop([idx_pu, idx_do], inplace=True)
            sl_s.drop([idx_pu, idx_do], inplace=True)
        else:
            sl.drop(idx_do, inplace=True)
            sl_s.drop([idx_pu, idx_do], inplace=True)

        res[
            f"{'system_' if scope=='system' else ''}stoplist_length_submission_time"
        ] = len(sl_s)
        res[f"{'system_' if scope=='system' else ''}stoplist_length_service_time"] = (
            len(sl)
        )

        res[
            f"avg_{'system_' if scope=='system' else ''}segment_dist_submission_time"
        ] = sl_s["dist_to_next"].mean()
        res[
            f"avg_{'system_' if scope=='system' else ''}segment_time_submission_time"
        ] = sl_s["time_to_next"].mean()

        res[f"avg_{'system_' if scope=='system' else ''}segment_dist_service_time"] = (
            sl["dist_to_next"].mean()
        )
        res[f"avg_{'system_' if scope=='system' else ''}segment_time_service_time"] = (
            sl["time_to_next"].mean()
        )

        return res

    stops = stops.merge(
        actual_stops.groupby("vehicle_id", group_keys=False).apply(
            lambda df: df.apply(
                ft.partial(_properties_at_time, full_sl=df, scope="vehicle"),
                axis=1,
                result_type="expand",
            )
        ),
        left_index=True,
        right_index=True,
        how="left",
    )

    stops = stops.merge(
        actual_stops.apply(
            ft.partial(_properties_at_time, full_sl=stops, scope="system"),
            axis=1,
            result_type="expand",
        ),
        left_index=True,
        right_index=True,
        how="left",
    )

    with pd.option_context("mode.use_inf_as_na", True):
        stops["relative_insertion_position"] = (
            stops["insertion_index"] / stops["stoplist_length_submission_time"]
        ).fillna(1)

    return stops



import pandas as pd
import numpy as np

import matplotlib.pyplot as plt


def plot_occupancy_hist(stops: pd.DataFrame, ax: plt.axes = None) -> plt.axes:
    """
    Create a occupancy histogram

    Parameters
    ----------
    stops
        stops dataframe
    ax
        axes object, optional

    Returns
    -------
    ax
        matplotlib axes

    """
    max_occ = max(stops.occupancy)
    if ax is None:
        _, ax = plt.subplots(figsize=(16, 8))

    ax.hist(
        stops["occupancy"],
        weights=stops["state_duration"],
        bins=np.r_[-0.5 : max_occ + 0.5 : 1j * (max_occ + 2)],
        rwidth=0.9,
    )
    ax.set_xticks(np.r_[0 : max_occ : 1j * (max_occ + 1)])
    ax.set_xlim((-1, max_occ + 1))
    ax.set_xlabel("occupancy")
    ax.set_ylabel("total duration at occupancy")
    ax.set_yscale("log")

    return ax



import networkx as nx
from typing import Tuple


def make_nx_grid(
    dim: Tuple[int, int] = (3, 3),
    periodic: bool = False,
    edge_distance: float = 1,
) -> nx.Graph:
    """
    Return a lattice `nx.Graph`

    Use in conjunction with `spaces.Graph` or `spaces_cython.Graph` like

    .. code-block:: python

        Graph.from_nx(make_nx_grid())

    Parameters
    ----------
    dim
        dimensions of the graph: (n, k) for n x k vertices
    periodic
        periodic boundaries
    edge_distance
        edge weight

    Returns
    -------
    graph
    """
    graph = nx.grid_graph(dim=dim, periodic=periodic)
    nx.set_edge_attributes(graph, edge_distance, "distance")
    graph = nx.relabel.convert_node_labels_to_integers(
        graph, label_attribute="location"
    )
    return graph


def make_nx_cycle_graph(order: int = 10, edge_distance: float = 1) -> nx.Graph:
    """
    Return a cyclic `nx.Graph`

    Use in conjunction with `spaces.Graph` or `spaces_cython.Graph` like

    .. code-block:: python

        Graph.from_nx(make_nx_cycle_graph())


    Parameters
    ----------
    order
        number of vertices to generate
    edge_distance
        edge weight

    Returns
    -------
    graph
    """
    graph = nx.generators.classic.cycle_graph(n=order)
    nx.set_edge_attributes(graph, edge_distance, "distance")
    return graph


def make_nx_star_graph(order: int = 10, edge_distance: float = 1) -> nx.Graph:
    """
    Return a cyclic `nx.Graph`

    Use in conjunction with `spaces.Graph` or `spaces_cython.Graph` like

    .. code-block:: python

        Graph.from_nx(make_nx_star_graph())

    Parameters
    ----------
    order
        number of vertices to generate in total (this includes the center node, which will have index 0)
    edge_distance
        edge weight

    Returns
    -------
    graph
    """
    graph = nx.generators.classic.star_graph(n=order - 1)
    nx.set_edge_attributes(graph, edge_distance, "distance")
    return graph



from collections.abc import Iterator

from numpy import inf
from abc import ABC, abstractmethod
from enum import Enum
from dataclasses import dataclass
from typing import Any, Optional, Union, Tuple, List, Callable


ID = Union[str, int]
"""Generic ID, could be vehicle ID, request ID, ..."""


@dataclass
class Request:
    """
    A request for the system to perform a task
    """

    request_id: ID
    creation_timestamp: float


@dataclass
class TransportationRequest(Request):
    """
    A request for the system to perform a transportation task,
    through creating a route through the system given spatio-temporal constraints.
    """

    origin: Any
    destination: Any
    # pickup_offset: float = 0
    pickup_timewindow_min: float = 0
    pickup_timewindow_max: float = inf
    delivery_timewindow_min: float = 0
    delivery_timewindow_max: float = inf


@dataclass
class InternalRequest(Request):
    """
    A request for the system to perform some action at a specific location
    that is not directly requested by a customer
    """

    location: Any


class StopAction(Enum):
    """
    Representing actions that the system may perform at a specific location
    """

    pickup = 1
    dropoff = 2
    internal = 3


class LocType(Enum):
    """
    Represents the kind of location objects the simulator supports. Either of:

    1. `R2LOC` (for points in :math:`\mathbb{R}^2`, holds a `Tuple[float, float]`).
    2. `INT` (for e.g. graphs).

    Note
    ----
    Use this for simulations using the pure python components. For simulations using cythonic components,
    the cython version of this enum i.e. :class:`.data_structures_cython.LocType` has to be used.
    """

    R2LOC = 1  # points in R^2
    INT = 2


@dataclass
class Stop:
    """
    The notion of an action to be performed in fulfilling a request.
    Attached are spatio-temporal constraints.

    Parameters
    ----------
    location:
        location at which the stop is supposed to be serviced
    """

    location: Any
    request: Request
    action: StopAction
    estimated_arrival_time: float
    occupancy_after_servicing: int = 0
    time_window_min: float = 0
    time_window_max: float = inf

    @property
    def estimated_departure_time(self):
        return max(
            self.estimated_arrival_time,
            self.time_window_min,
        )


class TransportSpace(ABC):
    @abstractmethod
    def d(self, u, v) -> Union[int, float]:
        """
        Return distance between points `u` and `v`.

        Parameters
        ----------
        u
            origin coordinate
        v
            destination coordinate

        Returns
        -------
        d
            distance
        """
        ...

    @abstractmethod
    def t(self, u, v) -> Union[int, float]:
        """
        Return travel time between points `u` and `v`.

        Parameters
        ----------
        u
            origin coordinate
        v
            destination coordinate

        Returns
        -------
        d
            travel time
        """

        ...

    @abstractmethod
    def random_point(self):
        """
        Return a random point on the space.

        Returns
        -------
            random point
        """
        ...

    @abstractmethod
    def interp_time(self, u, v, time_to_dest) -> Tuple[Any, Union[int, float]]:
        """
        Interpolate a location `x` between the origin `u` and the destination `v`
        as a function of the travel time between the unknown
        location and the destination `t(x, v) == time_to_dest`.

        Parameters
        ----------
        u
            origin coordinate
        v
            destination coordinate

        time_to_dest
            travel time from the unknown location `x` to the destination `v`

        Returns
        -------
        x
            interpolated coordinate of the unknown location `x`
        jump_dist
            remaining distance until the returned interpolated coordinate will be reached

        Note
        ----
        The notion of `jump_dist` is necessary in transport spaces whose locations are *discrete* (e.g. graphs). There
        if someone is travelling along a trajectory, at a certain time `t` one may be "in between" two locations `w` \
        and `x`. Then the "position" at time `t` is ill defined, and we must settle for the fact that its location
        *will be* `x` at `t+jump_time`.
        """
        ...

    @abstractmethod
    def interp_dist(
        self, origin, destination, dist_to_dest
    ) -> Tuple[Any, Union[int, float]]:
        """
        Interpolate a location `x` between the origin `u` and the destination `v`
        as a function of the distance between the unknown
        location and the destination `d(x, v) == dist_to_dest`.

        Parameters
        ----------
        u
            origin coordinate
        v
            destination coordinate

        dist_to_dest
            distance from the unknown location `x` to the destination `v`

        Returns
        -------
        x
            interpolated coordinate of the unknown location `x`
        jump_dist
            remaining distance until the returned interpolated coordinate will be reached
        """
        ...

    @abstractmethod
    def asdict(self) -> dict: ...

    def __eq__(self, other: "TransportSpace"):
        return type(self) == type(other) and self.asdict() == other.asdict()


class RequestGenerator(Iterator): ...


Stoplist = List[Stop]
"""A list of `.Stop` objects. Specifies completely the current position and future actions a vehicle will make."""

DispatcherSolution = tuple[float, Stoplist, tuple[float, float, float, float]]
"""cost, updated_stoplist, (
    pickup_timewindow_min,
    pickup_timewindow_max,
    delivery_timewindow_min, 
    delivery_timewindow_max,
)

This is what a dispatcher returns. In case no solution is found, the cost is 
:math:`\infty` and the timewindow variables are ``None``.
"""

SingleVehicleSolution = tuple[ID, float, tuple[float, float, float, float]]
"""vehicle_id, cost, (
    pickup_timewindow_min,
    pickup_timewindow_max,
    delivery_timewindow_min, 
    delivery_timewindow_max,
)

This is what `VehicleState.handle_transportation_request_single_vehicle` returns. 
In case no solution is found, the cost is :math:`\infty` and the timewindow variables are `None`.
"""

Dispatcher = Callable[
    [
        TransportationRequest,
        Stoplist,
        TransportSpace,
        int,
    ],
    DispatcherSolution,
]
"""Defines the `Dispatcher` interface. Actual dispatchers are implemented in `.util.dispatchers`."""

Location = Union[int, float, tuple[float]]


class DistanceDistribution(ABC):
    """
    Abstract base class for specifying distance distributions to use, e.g., during
    request generation.
    """

    def __init__(self, *args, **kwargs): ...

    @abstractmethod
    def sample(self) -> float:
        """
        Sample a distance from the distribution.
        """
        ...

    @property
    @abstractmethod
    def mean(self) -> float:
        """
        Return the mean of the distribution.
        """
        ...

    @property
    @abstractmethod
    def std(self) -> float:
        """
        Return the standard deviation of the distribution.
        """
        ...

    @abstractmethod
    def __repr__(self): ...

    @abstractmethod
    def __str__(self): ...

    @abstractmethod
    def asdict(self) -> dict[str, Any]:
        """
        Return a dictionary representation of the distribution.
        """
        ...
